import numpy as np
import pandas as pd
from scipy.stats import norm


def dprime(hit_rate=None, fa_rate=None, go_trials=None, catch_trials=None, limits=False):
    '''
    calculates the d-prime for a given hit rate and false alarm rate

    https://en.wikipedia.org/wiki/Sensitivity_index

    Parameters
    ----------
    hit_rate : float or vector of floats
        rate of hits in the True class
    fa_rate : float or vector of floats
        rate of false alarms in the False class
    go_trials: vector of booleans
        responses on all go trials (hit = True, miss = False)
    catch_trials: vector of booleans
        responses on all catch trials (false alarm = True, correct reject = False)
    limits : boolean or tuple, optional
        limits on extreme values, which can cause d' to overestimate on low trial counts.
        False (default) results in limits of (0.01,0.99) to avoid infinite calculations
        True results in limits being calculated based on trial count (only applicable if go_trials and catch_trials are passed)
        (limits[0], limits[1]) results in specified limits being applied

    Note: user must pass EITHER hit_rate and fa_rate OR go_trials and catch trials

    Returns
    -------
    d_prime

    Examples
    --------
    With hit and false alarm rates of 0 and 1, if we pass in limits of (0, 1) we are
    allowing the raw probabilities to be used in the dprime calculation.
    This will result in an infinite dprime:

    >> dprime(hit_rate = 1.0, fa_rate = 0.0, limits = (0, 1))
    np.inf

    If we do not pass limits, the default limits of 0.01 and 0.99 will be used
    which will convert the hit rate of 1 to 0.99 and the false alarm rate of 0 to 0.01.
    This will prevent the d' calcluation from being infinite:

    >>> dprime(hit_rate = 1.0, fa_rate = 0.0)
    4.6526957480816815

    If the hit and false alarm rates are already within the limits, the limits don't apply
    >>> dprime(hit_rate = 0.6, fa_rate = 0.4)
    0.5066942062715994

    Alternately, instead of passing in pre-computed hit and false alarm rates,
    we can pass in a vector of results on go-trials and catch-trials.
    Then, if we call `limits = True`, the limits will be calculated based
    on the number of trials in both the go and catch trial vectors
    using the `trial_number_limit` function.

    For example, for low trial counts, even perfect performance (hit rate = 1, false alarm rate = 0)
    leads to a lower estimate of d', given that we have low confidence in the hit and false alarm rates;

    >>> dprime(
            go_trials = [1, 1, 1],
            catch_trials = [0, 0],
            limits = True
            )
    1.6419113162977828

    At the limit, if we have only one sample of both trial types, the `trial_number_limit`
    pushes our estimated response probability to 0.5 for both the hit and false alarm rates,
    giving us a d' value of 0:

    >>> dprime(
            go_trials = [1],
            catch_trials = [0],
            limits = True
            )
    0.0

    And with higher trial counts, the `trial_number_limit` allows the hit and false alarm
    rates to get asymptotically closer to 0 and 1, leading to higher values of d'.
    For example, for 10 trials of each type:

    >>> dprime(
            go_trials = [1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
            catch_trials = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            limits = True
        )
    3.289707253902945

    Or, if we had 20 hit trials and 10 false alarm trials:

    >>> dprime(
            go_trials = [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
            catch_trials = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            limits = True
        )
    3.604817611491527

    Note also that Boolean vectors can be passed:

    >>> dprime(
            go_trials = [True, False, True, True, False],
            catch_trials = [False, True, False, False, False],
            limits = True
        )
    1.094968336708714
    '''

    assert hit_rate is not None or go_trials is not None, 'must either a specify `hit_rate` or pass a boolean vector of `go_trials`'
    assert fa_rate is not None or catch_trials is not None, 'must either a specify `fa_rate` or pass a boolean vector of `catch_trials`'

    assert hit_rate is None or go_trials is None, 'do not pass both `hit_rate` and a boolean vector of `go_trials`'
    assert fa_rate is None or catch_trials is None, 'do not pass both `fa_rate` and a boolean vector of `catch_trials`'

    assert not (hit_rate is not None and limits is True), 'limits can only be calculated if a go_trials vector is passed, not a hit_rate'
    assert not (fa_rate is not None and limits is True), 'limits can only be calculated if a catch_trials vector is passed, not a fa_rate'

    # calculate hit and fa rates as mean of boolean vectors
    if hit_rate is None:
        hit_rate = np.mean(go_trials)
    if fa_rate is None:
        fa_rate = np.mean(catch_trials)

    Z = norm.ppf

    if limits is False:
        # if limits are False, apply default
        limits = (0.01, 0.99)
    elif limits is True:
        # clip the hit and fa rate based on trial count
        hit_rate = response_probabilities_trial_number_limit(
            hit_rate, len(go_trials))
        fa_rate = response_probabilities_trial_number_limit(
            fa_rate, len(catch_trials))

    if limits is not True:
        # Limit values in order to avoid d' infinity
        hit_rate = np.clip(hit_rate, limits[0], limits[1])
        fa_rate = np.clip(fa_rate, limits[0], limits[1])

    # keep track of nan locations
    hit_rate = pd.Series(hit_rate)
    fa_rate = pd.Series(fa_rate)
    hit_rate_nan_locs = list(hit_rate[pd.isnull(hit_rate)].index)
    fa_rate_nan_locs = list(fa_rate[pd.isnull(fa_rate)].index)

    # fill nans with 0.0 to avoid warning about nans
    d_prime = Z(hit_rate.fillna(0)) - Z(fa_rate.fillna(0))

    # for every location in hit_rate and fa_rate with a nan, fill d_prime with a nan
    for nan_locs in [hit_rate_nan_locs, fa_rate_nan_locs]:
        d_prime[nan_locs] = np.nan

    if len(d_prime) == 1:
        # if the result is a 1-length vector, return as a scalar
        return d_prime[0]
    else:
        return d_prime


def response_probabilities_trial_number_limit(P, N):
    '''
    Calculates limits on response probability estimate based on trial count.
    An important point to note about d' is that the metric will be infinite with
    perfect performance, given that Z(0) = -infinity and Z(1) = infinity.
    Low trial counts exacerbate this issue. For example, with only one sample of a go-trial,
    the hit rate will either be 1 or 0. Macmillan and Creelman [1] offer a correction on the hit and false
    alarm rates to avoid infinite values whereby the response probabilities (P) are bounded by
    functions of the trial count, N:

        1/(2N) < P < 1 - 1/(2N)

    Thus, for the example of just a single trial, the trial-corrected hit rate would be 0.5.
    Or after only two trials, the hit rate could take on the values of 0.25, 0.5, or 0.75.

    [1] Macmillan, Neil A., and C. Douglas Creelman. Detection theory: A user's guide. Psychology press, 2004.
    Parameters
    ----------
    P : float
        response probability, bounded by 0 and 1
    N : int
        number of trials used in the response probability calculation

    Returns
    -------
    P_corrected : float
        The response probability after adjusting for the trial count

    Examples
    --------
    Passing in a response probability of 1 and trial count of 10 will lead to a reduced
    estimate of the response probability:

    >>> trial_number_limit(1, 10)
    0.95

    A higher trial count increases the bounds on the estimate:

    >>> trial_number_limit(1, 50)
    0.99

    At the limit, the bounds are 0 and 1:

    >>> trial_number_limit(1, np.inf)
    1.0

    The bounds apply for estimates near 0 also:

    >>> trial_number_limit(0, 1)
    0.5

    >>> trial_number_limit(0, 2)
    0.25

    >>> trial_number_limit(0, 3)
    0.16666666666666666

    Note that passing a probality that is inside of the bounds
    results in no change to the passed probability:

    >>> trial_number_limit(0.25, 3)
    0.25
    '''
    if N == 0:
        return np.nan
    if not pd.isnull(P):
        P = np.max((P, 1. / (2 * N)))
        P = np.min((P, 1 - 1. / (2 * N)))
    return P



# Plotting style enforcment dictionaries
# to provide colors and labels for different
# data dimensions. Colors are all hexidecimel


DATASTREAM_STYLE_DICT = {
    "licks":                  {"color": "#252525",
                               "label": "licks"},
    "rewards":                {"color": "#6baed6",
                               "label": "rewards"},
    "pupil_area":             {"color": "#fdc086",
                               "label": "pupil area"},
    "dff":                    {"color": "#7fc97f",
                               "label":  "df/f"},
    "running_speed":          {"color": "#beaed4",
                               "label": "running speed (cm/sec)"},
    "stimulus_presentations": {"color": "#cccccc",
                               "label": "stim presentations"}
}
REWARDS_STYLE_DICT = {
    "all":    {"color": "#6baed6", "label": "rewards (all)"},
    "earned": {"color": "#08519c", "label": "rewards (earned)"},
    "auto":   {"color": "#c6dbef", "label": "rewards (auto)"}
}
CRELINE_STYLE_DICT = {
    "Sst-IRES-Cre":      {"color": "#b35806", "label": "Sst"},
    "Vip-IRES-Cre":      {"color": "#fdb863", "label": "Vip"},
    "Slc17a7-IRES2-Cre": {"color": "#8073ac", "label": "Slc"}
}

NOVELTY_STYLE_DICT = {
    "TRAINING_0": {"color": "#525252", "label": "training: gratings"},
    "TRAINING_1": {"color": "#525252", "label": "training: gratings"},
    "TRAINING_2": {"color": "#525252", "label": "training: gratings"},
    "TRAINING_3": {"color": "#bdbdbd", "label": "training: images"},
    "TRAINING_4": {"color": "#bdbdbd", "label": "training: images"},
    "TRAINING_5": {"color": "#bdbdbd", "label": "training: images"},
    "OPHYS_0":    {"color": "#f7f7f7", "label": "habituation"},
    "OPHYS_1":    {"color": "#2166ac", "label": "familiar 1"},
    "OPHYS_2":    {"color": "#67a9cf", "label": "familiar 2"},
    "OPHYS_3":    {"color": "#d1e5f0", "label": "familiar 3"},
    "OPHYS_4":    {"color": "#b2182b", "label": "novel 1"},
    "OPHYS_5":    {"color": "#ef8a62", "label": "novel 2"},
    "OPHYS_6":    {"color": "#fddbc7", "label": "novel 3"}
}

TRAINING_STYLE_DICT = {
    "TRAINING_0": {"color": "#525252", "label": "TRAINING 0 gratings"},
    "TRAINING_1": {"color": "#737373", "label": "TRAINING 1 gratings"},
    "TRAINING_2": {"color": "#969696", "label": "TRAINING 2 gratings"},
    "TRAINING_3": {"color": "#bdbdbd", "label": "TRAINING 3 images"},
    "TRAINING_4": {"color": "#d9d9d9", "label": "TRAINING 4 images"},
    "TRAINING_5": {"color": "#f7f7f7", "label": "TRAINING 5 images"}
}


IMAGE_SET_STYLE_DICT = {
    "A": {"color": "#a6611a", "label": "A"},
    "B": {"color": "#018571", "label": "B"},
    "G": {"color": "#dfc27d", "label": "G"},
    "H": {"color": "#80cdc1", "label": "H"},
    "gratings": {"color": "#737373", "label": "gratings"}
}

BEHAV_RESP_STYLE_DICT = {
    "hit":            {"color": "#4dac26", "label": "hit"},
    "miss":           {"color": "#d01c8b", "label": "miss"},
    "false_alarm":    {"color": "#f1b6da", "label": "false alarm"},
    "correct_reject": {"color": "#b8e186", "label": "correct reject"},
    "aborted":        {"color": "#525252", "label": "aborted"}
}


def get_color_for_dict_key(dictionary, key):
    color = dictionary[key]['color']
    return color


def get_label_for_dict_key(dictionary, key):
    label = dictionary[key]['label']
    return label


def get_style_for_stim_name(dictionary, stimulus_name):
    dict_keys = dictionary.keys()
    for key in dict_keys:
        if key in stimulus_name:
            return dictionary[key]["color"], dictionary[key]["label"]
        else:
            print("Stimulus name cannot be aligned dictionary keys.")


def get_image_set_style_for_stim_name(stimulus_name):
    return get_style_for_stim_name(IMAGE_SET_STYLE_DICT,
                                   stimulus_name)


def get_novelty_style_for_stim_name(stimulus_name):
    return get_style_for_stim_name(NOVELTY_STYLE_DICT,
                                   stimulus_name)



import numpy as np
import pandas as pd
import brain_observatory_utilities.datasets.optical_physiology.data_formatting as ophys


def get_stimulus_block_description(stimulus_block, dataset='VBN'):
    """
    For a given dataset, will return a string description of what was shown during the provided stimulus_block.
    Currently only applies to dataset='VBN' because stimulus_block has not yet been applied to VBO

    :param stimulus_block: int between 0 and 5
    :param dataset: 'VBN' = Visual Behavior Neuropixels, 'VBO' = Visual Behavior Ophys
    :return: string description of what stimulus was shown during the provided stimulus_block
    """
    if dataset == 'VBO':
        print('stimulus block is not currently used in the Visual Behavior Ophys dataset')
        stimulus_block_lookup_dict = {}
    elif dataset == 'VBN':
        stimulus_block_lookup_dict = {0: 'change detection active behavior',
                                      1: '10 seconds gray screen',
                                      2: 'gabor patches receptive field mapping',
                                      3: '5 minutes gray screen',
                                      4: 'full field flashes',
                                      5: 'change detection passive replay'}
        print(stimulus_block_lookup_dict[stimulus_block], 'is shown during Visual Behavior Neuropixels stimulus_block', stimulus_block)
    return stimulus_block_lookup_dict[stimulus_block]


def get_stimulus_name(dataObject):
    """gets the stimulus name for a dataset object
    Parameters
    ----------
    dataObject : (BehaviorSesson, BehaviorOphysExperiment)
        Objects provided via allensdk.brain_observatory module

    Returns
    -------
    string
        the stimulus name for a given session or experiment
    """
    stimulus_name = dataObject.metadata['session_type']
    return stimulus_name


def get_lick_timestamps(dataObject):
    """gets the timestamps of licks

    Parameters
    ----------
    dataObject : (BehaviorSesson, BehaviorOphysExperiment)
        Objects provided via allensdk.brain_observatory module
    Returns
    -------
    array
        array of lick timestamps
    """
    lick_timestamps = dataObject.licks["timestamps"].values
    return lick_timestamps


def get_reward_timestamps(dataObject, reward_type="all"):
    """gets the timestamps of water rewards
    Parameters
    ----------
    dataObject : (BehaviorSesson, BehaviorOphysExperiment)
        Objects provided via allensdk.brain_observatory module
    reward_type : string
        by default "all", options:
            "all": all rewards (auto and earned)
            "auto": only free or unearned rewards
            "earned": only earned (hit on a go trial) rewards

    Returns
    -------
    array
        arry of reward timestamps
    """

    rewards_df = dataObject.rewards
    if reward_type == "all":
        reward_timestamps = rewards_df['timestamps'].values
    elif reward_type == "auto":
        reward_timestamps = rewards_df.loc[rewards_df["autorewarded"] == True]["timestamps"].values  # noqa: E501
    elif reward_type == "earned":
        reward_timestamps = rewards_df.loc[rewards_df["autorewarded"] == False]["timestamps"].values  # noqa: E501
    return reward_timestamps


def get_running_speed_timeseries(dataObject):
    """gets the mouse running speed timeseries

    Parameters
    ----------
    ophysObject : (BehaviorSesson, BehaviorOphysExperiment)
        Objects provided via allensdk.brain_observatory module

    Returns
    -------
    tuple:
        running_speed (cm/sec), timestamps (sec)
    """
    running_speed = dataObject.running_speed["speed"].values
    timestamps = dataObject.running_speed["timestamps"].values
    return running_speed, timestamps


def get_pupil_area_timeseries(ophysObject):
    """gets mouse's pupil area timeseries

    Parameters
    ----------
    ophysObject : (BehaviorOphysExperiment)
        Object provided via allensdk.brain_observatory
        module

    Returns
    -------
   tuple
        pupil area (pixels ^2), timestamps
    """
    pupil_area = ophysObject.eye_tracking["pupil_area"].values
    timestamps = ophysObject.eye_tracking["timestamps"].values
    return pupil_area, timestamps


def get_trial_type(trials_df, trial_type,
                   include_aborted):
    """filters the trials table to a specific trial type

    Parameters
    ----------
    trials_df : pandas dataframe
        _description_
    trial_type : str
        options:
            "all": all trial types (both go and catch)
            "go": go trials (change presented)
            "catch": catch trials (no change presented)
    include_aborted : bool,
        include aborted trials,
    Returns
    -------
    pandas dataframe
        filtered trials table
    """
    if trial_type == "all":
        if include_aborted is False:
            filtered_trials = trials_df.loc[trials_df["aborted"] == False]  # noqa: E501
        else:
            filtered_trials = trials_df
    else:
        filtered_trials = \
            trials_df.loc[trials_df[trial_type] == True]
    return filtered_trials


def get_response_type(trials_df, response_type,
                      include_aborted):
    """filters the trials table to a specific mouse
       behavior response type.

    Parameters
    ----------
    trials_df : pandas dataframe
        _description_
    response_type : str,
        options:
            "all": all responses
            "hit": "hit" responses, only occurs on go trial type
            "miss": "miss" responses, only occurs on
                go trial type
            "false_alarm": "false_alarm" responses,
                only occurs on catch trial type
            "correct_reject": "correct_reject" responses,
                only occurs on catch trial type
    include_aborted : bool
        include aborted trials, by default True

    Returns
    -------
    pandas dataframe
        filtered trials table
    """
    if response_type == "all":
        if include_aborted is False:
            filtered_trials = trials_df.loc[trials_df["aborted"] == False]
        else:
            filtered_trials = trials_df
    else:
        filtered_trials = trials_df.loc[trials_df[response_type] == True]
    return filtered_trials


def get_trials_data(dataObject,
                    trial_type="all", response_type="all",
                    include_aborted_trials=True):
    """gets the trials dataframe attribute and can optionally
       filter for specific trial or behavior response types.

    Parameters
    ----------
    dataObject : (BehaviorSesson, BehaviorOphysExperiment)
        Objects provided via allensdk.brain_observatory module
    trial_type : str, optional
        by default "all", options:
            "all": all trial types (both go and catch)
            "go": go trials (change presented)
            "catch": catch trials (no change presented)
    response_type : str, optional
        by default "all", options:
            "all": all responses
            "hit": "hit" responses, only occurs on go trial type
            "miss": "miss" responses, only occurs on
                go trial type
            "false_alarm": "false_alarm" responses,
                only occurs on catch trial type
            "correct_reject": "correct_reject" responses,
                only occurs on catch trial type
    include_aborted_trials : bool, optional
        whether to include aborted trials, by default True

    Returns
    -------
    pandas dataframe
        trials dataframe, optionally filtered
    """
    trials_df = dataObject.trials
    filtered_df = get_trial_type(trials_df,
                                 trial_type=trial_type,
                                 include_aborted=include_aborted_trials)
    filtered_df = get_response_type(filtered_df,
                                    response_type=response_type,
                                    include_aborted=include_aborted_trials)
    return filtered_df


def get_pupil_data(eye_tracking, interpolate_likely_blinks=False, normalize_to_gray_screen=False, zscore=False,
                   interpolate_to_ophys=False, ophys_timestamps=None, stimulus_presentations=None):
    """
    removes 'likely_blinks' from all columns in eye_tracking and converts pupil_area to pupil_diameter and pupil_radius
    interpolates over NaNs resulting from removing likely_blinks if interpolate = true
    :param eye_tracking: eye_tracking attribute of AllenSDK BehaviorOphysExperiment object
    :param column_to_use: 'pupil_area', 'pupil_width', 'pupil_diameter', or 'pupil_radius'
                            'pupil_area' and 'pupil_width' are existing columns of the eye_tracking table
                            'pupil_diameter' and 'pupil_radius' are computed from 'pupil_area' column
    :param interpolate: Boolean, whether or not to interpolate points where likely_blinks occured
    """


    # compute pupil_diameter and pupil_radius from pupil_area
    # convert pupil area to pupil diameter
    eye_tracking['pupil_diameter'] = np.sqrt(eye_tracking.pupil_area) / np.pi
    eye_tracking['pupil_radius'] = np.sqrt(
        eye_tracking['pupil_area'] * (1 / np.pi))  # convert pupil area to pupil radius

    # set all timepoints that are likely blinks to NaN for all eye_tracking columns
    if True in eye_tracking.likely_blink.unique():  # only can do this if there are likely blinks to filter out
        eye_tracking.loc[eye_tracking['likely_blink'], :] = np.nan

    # add timestamps column back in
    eye_tracking['timestamps'] = eye_tracking.index.values

    # interpolate over likely blinks, which are now NaNs
    if interpolate_likely_blinks:
        eye_tracking = eye_tracking.interpolate()

    # divide all columns by average of gray screen period prior to behavior session
    if normalize_to_gray_screen:
        assert stimulus_presentations is not None, 'must provide stimulus_presentations if normalize_to_gray_screen is True'
        spontaneous_frames = ophys.get_spontaneous_frames(stimulus_presentations,
                                                          eye_tracking.timestamps.values,
                                                          gray_screen_period_to_use='before')
        for column in eye_tracking.keys():
            if (column != 'timestamps') and (column != 'likely_blink'):
                gray_screen_mean_value = np.nanmean(
                    eye_tracking[column].values[spontaneous_frames])
                eye_tracking[column] = eye_tracking[column] / \
                    gray_screen_mean_value
    # z-score pupil data
    if zscore:
        for column in eye_tracking.keys():
            if (column != 'timestamps') and (column != 'likely_blink'):
                eye_tracking[column] = scipy.stats.zscore(
                    eye_tracking[column], nan_policy='omit')

    # interpolate to ophys timestamps
    if interpolate_to_ophys:
        assert ophys_timestamps is not None, 'must provide ophys_timestamps if interpolate_to_ophys is True'
        eye_tracking_ophys_time = pd.DataFrame(
            {'timestamps': ophys_timestamps})
        for column in eye_tracking.keys():
            if (column != 'timestamps') and (column != 'likely_blink'):
                f = scipy.interpolate.interp1d(
                    eye_tracking['timestamps'], eye_tracking[column], bounds_error=False)
                eye_tracking_ophys_time[column] = f(
                    eye_tracking_ophys_time['timestamps'])
                eye_tracking_ophys_time[column].fillna(
                    method='ffill', inplace=True)
        eye_tracking = eye_tracking_ophys_time

    return eye_tracking



import numpy as np
import pandas as pd


def makePSTH(spikes, startTimes, windowDur, binSize=0.001):
    '''
    Convenience function to compute a peri-stimulus-time histogram
    (see section 7.2.2 here: https://neuronaldynamics.epfl.ch/online/Ch7.S2.html)
    INPUTS:
        spikes: spike times in seconds for one unit
        startTimes: trial start times in seconds; the first spike count 
            bin will be aligned to these times
        windowDur: trial duration in seconds
        binSize: size of spike count bins in seconds
    OUTPUTS:
        Tuple of (PSTH, bins), where:
            PSTH gives the trial-averaged spike rate for 
                each time bin aligned to the start times;
            bins are the bin edges as defined by numpy histogram
    '''
    bins = np.arange(0, windowDur+binSize, binSize)
    counts = np.zeros(bins.size-1)
    for start in startTimes:
        startInd = np.searchsorted(spikes, start)
        endInd = np.searchsorted(spikes, start+windowDur)
        counts = counts + np.histogram(spikes[startInd:endInd]-start, bins)[0]
    
    counts = counts/len(startTimes)
    return counts/binSize, bins[:-1]


def make_psth(spike_times, stim_times, pre_window=0.5, post_window=1.0, bin_size=0.05):
    """
    Generate a Peri-Stimulus Time Histogram (PSTH).
    
    Parameters:
    - spike_times: array-like, timestamps of all spikes (in seconds)
    - stim_times: array-like, timestamps of stimulus onsets (in seconds)
    - pre_window: float, time before stimulus to include in PSTH (seconds)
    - post_window: float, time after stimulus to include in PSTH (seconds)
    - bin_size: float, width of each time bin (seconds)
    
    Returns:
    - firing_rates: 2D numpy array of firing rates (trials x bins)
    - bin_centers: 1D numpy array of bin center times (relative to stimulus onset)
    """

    # Ensure inputs are numpy arrays 
    spike_times = np.array(spike_times)
    stim_times = np.array(stim_times)
    
    # Define bin edges from -pre_window to +post_window
    bins = np.arange(-pre_window, post_window + bin_size, bin_size)
    
    # Compute centers of bins (for plotting)
    bin_centers = bins[:-1] + bin_size / 2
    
    # Initialize a matrix to hold spike counts: rows = trials, columns = bins
    all_counts = np.zeros((len(stim_times), len(bins) - 1))
    
    # Loop through each stimulus time to compute trial-specific spike counts
    for i, stim_time in enumerate(stim_times):
        # Select spikes that fall within the time window around this stimulus
        mask = ((spike_times >= stim_time - pre_window) & 
                (spike_times < stim_time + post_window))
        
        # Align spike times to stimulus onset (0 = stimulus)
        trial_spikes = spike_times[mask] - stim_time
        
        # Bin the aligned spikes and count how many fall into each bin
        counts, _ = np.histogram(trial_spikes, bins=bins)
        
        # Store the result in the i-th row (trial)
        all_counts[i, :] = counts
    
    # Convert spike counts to firing rates (spikes per second)
    firing_rates = all_counts / bin_size
    
    # Return firing rates (trials x bins) and bin center positions
    return firing_rates, bin_centers


def make_neuron_time_trials_array(units, spike_times, stim_table, 
                                   time_before, trial_duration,
                                   bin_size=0.001):
    '''
    Function to make a 3D array with dimensions [neurons, time bins, trials] to store
    the spike counts for stimulus presentation trials. 
    INPUTS:
        units: dataframe with unit info (same form as session.units table)
        spike_times: dictionary with spike times for each unit (ie session.spike_times)
        stim_table: dataframe whose indices are trial ids and containing a
            'start_time' column indicating when each trial began
        time_before: seconds to take before each start_time in the stim_table
        trial_duration: total time in seconds to take for each trial
        bin_size: bin_size in seconds used to bin spike counts 
    OUTPUTS:
        unit_array: 3D array storing spike counts. The value in [i,j,k] 
            is the spike count for neuron i at time bin j in the kth trial.
        time_vector: vector storing the trial timestamps for the time bins
    '''
    # Get dimensions of output array
    neuron_number = len(units)
    trial_number = len(stim_table)
    num_time_bins = int(trial_duration/bin_size)
    
    # Initialize array
    unit_array = np.zeros((neuron_number, num_time_bins, trial_number))
    
    # Loop through units and trials and store spike counts for every time bin
    for u_counter, (iu, unit) in enumerate(units.iterrows()):
        
        # grab spike times for this unit
        unit_spike_times = spike_times[iu]
        
        # now loop through trials and make a PSTH for this unit for every trial
        for t_counter, (it, trial) in enumerate(stim_table.iterrows()):
            trial_start = trial.start_time - time_before
            unit_array[u_counter, :, t_counter] = makePSTH(unit_spike_times, 
                                                            [trial_start], 
                                                            trial_duration, 
                                                            binSize=bin_size)[0]
    
    # Make the time vector that will label the time axis
    time_vector = np.arange(num_time_bins)*bin_size - time_before
    
    return unit_array, time_vector


def get_good_units(session):
    # get units table
    this_session_units = session.get_units()

    # Apply QC criteria
    good_units = this_session_units[
        (this_session_units.isi_violations<.5) &
        (this_session_units.amplitude_cutoff<.1) &
        (this_session_units.presence_ratio>.95)]
    print(len(good_units), 'units passing QC criteria')

    return good_units


def get_continous_spike_rate_for_units(session, stimulus_presentations, spike_rate_bin_size=0.01):
    """
    Create dataframe containing continuous spike rates across a full stimulus block for all units

    session: SDK VBN session object
    stimulus_presentations: stimulus_presentations table limited to the stimuli or stimulus blocks of interest
    spike_rate_bin_size: bin size, in seconds, to use when computing spike rate over time
                0.001 = 1ms, 0.01 = 10ms, 1 = 1s (spikes / second)
    stimulus_block: stimulus block number indicating portion of VBN session to compute spike rates for
                stimulus block 0 = change detection active behavior, 1 = 10s gray screen, 2 = gabor RF mapping,
                3 = 5min gray screen, 4 = full field flashes, 5 = change detection passive replay
    """

    # get data
    units = get_good_units(session)
    spike_times = session.spike_times.copy()
    stim_table = stimulus_presentations.copy()
    # create timeframe from start to end of behavior block
    start_time = stim_table.start_time.values[0]
    end_time = stim_table.end_time.values[-1] + 10 # Add 10 seconds to make sure the last image flashes can be included
    behavior_duration = end_time - start_time

    # Get dimensions of output array
    unit_ids = units.index.values
    neuron_number = len(unit_ids)
    num_time_bins = int(behavior_duration / spike_rate_bin_size) + 1

    # Initialize array
    unit_array = np.zeros((neuron_number, num_time_bins))

    # Loop through units and trials and store spike counts for every unit
    for i, unit_id in enumerate(unit_ids):
        # grab spike times for this unit
        unit_spike_times = spike_times[unit_id]
        spike_rate, timestamps = make_psth(unit_spike_times, [start_time], pre_window=0, 
                                           post_window=behavior_duration, bin_size=spike_rate_bin_size)
        # spike_rate, timestamps = makePSTH(unit_spike_times, [start_time], behavior_duration, binSize=spike_rate_bin_size)
        # convert to spikes per second
        spike_rate = spike_rate * spike_rate_bin_size
        unit_array[i, :] = spike_rate
    
    # turn it into a df where each row is a uit and column contains entire spike rate trace
    # to match format of ophys dff_traces table
    spike_rate_df = pd.DataFrame(index=units.index, columns=['spike_rate'])
    for i, unit_id in enumerate(unit_ids):
        spike_rate_df.loc[unit_id, 'spike_rate'] = unit_array[i, :]
    spike_rate_df.index.name = 'unit_id'

    return spike_rate_df, timestamps+start_time


def build_tidy_cell_df(dataset, stimulus_presentations, spike_rate_bin_size=0.01):
    '''
    Builds a tidy dataframe describing activity for every unit in ephys session.
    Tidy format is defined as one row per observation.
    Thus, the output dataframe will be n_units x n_timepoints long

    Parameters:
    -----------
    dataset : AllenSDK BehaviorEcephysSession object
    stimulus_presentations: stimulus_presentations table limited to the stimuli or stimulus blocks of interest
    spike_rate_bin_size: bin size, in seconds, to use when computing spike rate over time
                0.001 = 1ms, 0.01 = 10ms, 1 = 1s (spikes / second)
    stimulus_block: stimulus block number indicating portion of VBN session to compute spike rates for
                stimulus block 0 = change detection active behavior, 1 = 10s gray screen, 2 = gabor RF mapping,
                3 = 5min gray screen, 4 = full field flashes, 5 = change detection passive replay

    Returns:
    --------
    Pandas.DataFrame
        Tidy Format (one observation per row) with the following columns:
            * timestamps (float) : the ephys timestamps for the session, computed based on bin_size
            * unit_id (int) : ecephys unit ID
            * spike_rate (float) : measured spike rate for every timestep
    '''

    spike_rate_df, timestamps = get_continous_spike_rate_for_units(dataset, stimulus_presentations, spike_rate_bin_size)
    print('spike rate df computed')
    # make an empty list to populate with dataframes for each cell
    list_of_cell_dfs = []

    # iterate over each individual unit
    for unit_id in spike_rate_df.index.values:
        # build a tidy dataframe for this unit
        cell_df = pd.DataFrame({'timestamps': timestamps,
                                'spike_rate': spike_rate_df.loc[unit_id]['spike_rate']})  # noqa E501

        # Make the unit_id column categorical
        # This will reduce memory useage since the columns
        # consist of many repeated values.
        cell_df['unit_id'] = np.int32(unit_id)
        cell_df['unit_id'] = pd.Categorical(cell_df['unit_id'], categories=spike_rate_df.index.unique())

        # append the dataframe for this cell to the list of cell dataframes
        list_of_cell_dfs.append(cell_df)

    # concatenate all dataframes in the list
    tidy_df = pd.concat(list_of_cell_dfs)
    print('tidy cell df computed')

    # return the tidy dataframe
    return tidy_df


def getImageNovelty(image_name, session_id, ecephys_sessions_table):
    '''
    Function to help annotate the stimulus_presentations table
    to indicate whether the image was novel to the mouse.      
    
    INPUT:
        image_name: str indicating which image to check for novelty (ie 'im024_r')
        session_id: the ecephys_session_id for session during which image was presented
        ecephys_sessions_table: the ecephys_sessions metadata table from the VBN cache
    
    OUTPUT:
        Returns one of the following:
        True: indicating that this image was novel for this session
        False: indicating that this image was familiar for this session
        np.nan: indicating that this stimulus wasn't one of the natural images (including omitted stimuli)

    '''
    is_novel_image_set = ecephys_sessions_table.loc[session_id]['experience_level'] == 'Novel'

    IMAGE_SET_KEY={
                'G' : ['im012_r', 'im036_r', 'im044_r', 
                    'im047_r', 'im078_r', 'im115_r'],
                'H' : ['im005_r', 'im024_r', 'im034_r', 
                    'im087_r', 'im104_r', 'im114_r'],
                'shared' : ['im083_r', 'im111_r'],
                'omitted' : 'omitted'
                }
    
    # First check that this image is one of the Natural Images used
    image_in_image_set = any([np.isin(image_name, imset) \
                              for _,imset in IMAGE_SET_KEY.items()]) 
    
    if not image_in_image_set:
        return np.nan
    
    #Get the image set for this image
    image_set_for_this_image = [name for name, image_set in IMAGE_SET_KEY.items()\
                                if image_name in image_set][0]
    
    #Get the image novelty for this image
    if image_set_for_this_image == 'omitted':
        novelty_for_this_image = np.nan
    else:
        novelty_for_this_image = is_novel_image_set and \
                            bool(np.isin(image_set_for_this_image, ['G', 'H']))

    return novelty_for_this_image


import os
import sys
import numpy as np
import matplotlib.pyplot as plt

from .plot_utils import *

sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
import data_access as data  # noqa: 402


def plot_max_projection(ophysObject, ax=None):
    """plots the maximum intensity projection
    for the given dataset

    Parameters
    ----------
    ophysObject : BehaviorOphysExperiment
        Object provided via allensdk.brain_observatory
        module
    ax : matplotlib.axes, optional
        Figure axes

    Returns
    -------
    matplotlib.axes
        plotting axes
    """
    if ax is None:
        fig, ax = plt.subplots()
    ax.imshow(ophysObject.max_projection, cmap='gray')
    return ax


def plot_average_projection(ophysObject, ax=None):
    """plots the average intensity projection
    for the given dataset

    Parameters
    ----------
    ophysObject : BehaviorOphysExperiment
        Object provided via allensdk.brain_observatory
        module
    ax : matplotlib.axes, optional
        Figure axes

    Returns
    -------
    matplotlib.axes
        plotting axes
    """
    if ax is None:
        fig, ax = plt.subplots()
    ax.imshow(ophysObject.average_projection, cmap='gray')
    return ax


def plot_segmentation_masks(ophysObject, ax=None):
    """plots the masks of the segmented rois/cells

    Parameters
    ----------
    ophysObject : BehaviorOphysExperiment
        Object provided via allensdk.brain_observatory
        module
    ax : matplotlib.axes, optional
        Figure axes

    Returns
    -------
    matplotlib.axes
        plotting axes
    """
    if ax is None:
        fig, ax = plt.subplots()
    ax.imshow(ophysObject.segmentation_mask_image)
    return ax


def plot_transparent_segmentation_masks(ophysObject, ax=None):
    """plots the segmentation masks with a transparent
    background

    Parameters
    ----------
    ophysObject : BehaviorOphysExperiment
        Object provided via allensdk.brain_observatory
        module
    ax : matplotlib.axes, optional
        Figure axes

    Returns
    -------
    matplotlib.axes
        plotting axes
    """
    if ax is None:
        fig, ax = plt.subplots()
    transparent_mask = data.get_transparent_segmentation_mask(ophysObject)  # noqa: E501
    ax.imshow(transparent_mask, cmap='hsv', vmax=1, alpha=0.5)
    return ax


def plot_segmentation_mask_overlay(ophysObject, projection_type="max",
                                   ax=None):
    """plots the segmentation masks on top of the one of the
    projections (average intensity projection,
    max intensity projection)

    Parameters
    ----------
    ophysObject : BehaviorOphysExperiment
        Object provided via allensdk.brain_observatory
        module
    projection_type : str, optional
        by default "max",  options:
            "max" : maximum intensity projection
            "average" : average intensity projection
    ax : matplotlib.axes, optional
        Figure axes

    Returns
    -------
    matplotlib.axes
        plotting axes
    """

    if ax is None:
        fig, ax = plt.subplots()
    if projection_type == "max":
        ax = plot_max_projection(ophysObject, ax)
    elif projection_type == "average":
        ax = plot_average_projection(ophysObject, ax)
    else:
        print("Please enter a valid projection_type, \
            'max' or 'average'.")

    ax = plot_transparent_segmentation_masks(ophysObject, ax)
    ax.axis('off')
    return ax


def plot_dff(ophysObject, cell_specimen_id=None, ax=None):
    """ plots the fluroescence trace (dF/F) trace.
    x axis is time (sec) and y axis is fluorescence
    values

    Parameters
    ----------
    ophysObject :BehaviorOphysExperiment
        Object provided via allensdk.brain_observatory
        module
    cell_specimen_id : int, optional
        _description_, by default None
    ax : matplotlib.axes, optional
        Figure axes

    Returns
    -------
    matplotlib.axes
        plotting axes
    """
    if ax is None:
        fig, ax = plt.subplots()
        ax.set_xlabel("time (sec)")
    dff_trace, timestamps = data.get_dff_trace_timeseries(ophysObject, cell_specimen_id)  # noqa: E501
    ax.plot(timestamps, dff_trace, color=DATASTREAM_STYLE_DICT['dff']['color'])  # noqa: E501
    ax.set_title("Fluorescence trace")
    ax.set_ylabel("df/f")
    return ax


def plot_dff_heatmap(ophysObject, ax=None):
    """plots a heatmap of fluorescence activity for
    all cells in a given ophys experiment.

    Parameters
    ----------
    ophysObject : BehaviorOphysExperiment
        Object provided via allensdk.brain_observatory
        module
    ax : matplotlib.axes, optional
        Figure axes
    """
    dff_traces = data.get_all_cells_dff(ophysObject.dff_traces)
    timestamps = ophysObject.ophys_timestamps

    if ax is None:
        fig, ax = plt.subplots()

    fig, ax = plt.subplots(figsize=(20, 5))
    color_ax = ax.pcolormesh(dff_traces,
                             vmin=0, vmax=np.percentile(dff_traces, 99),
                             cmap='magma')
    # label axes
    ax.set_ylabel('cells')
    ax.set_xlabel('time (sec)')

    # x ticks
    ax.set_yticks(np.arange(0, len(dff_traces), 10))
    ax.set_xticklabels(np.arange(0, timestamps[-1], 300))

    # creating a color bar
    cb = plt.colorbar(color_ax, pad=0.015, label='dF/F')
    return ax




import numpy as np
from brain_observatory_utilities.datasets.behavior.data_access import *
import brain_observatory_utilities.utilities.general_utilities as gen_utils



def get_dff_trace_timeseries(ophysObject, cell_specimen_id=None):
    """ By default will return the average dff trace (mean
        of all cell dff traces) for an ophys experiment. If
        cell_specimen_id is specified then will return the
        dff trace for that single cell specimen id.

    Parameters
    ----------
    ophysObject : (BehaviorOphysExperiment)
        Object provided via allensdk.brain_observatory
        module
    cell_specimen_id : int
        unified id of segmented cell across experiments
        (assigned after cell matching). Will return dff
        trace for this single cell_specimen_id.

    Returns
    -------
    tuple
        dff_trace, timestamps (sec)
    """
    dff_traces_df = ophysObject.dff_traces.reset_index()
    if cell_specimen_id is None:
        dff = get_all_cells_mean_dff(dff_traces_df)
    else:
        dff = get_cell_dff(dff_traces_df, cell_specimen_id)
    timestamps = ophysObject.ophys_timestamps
    return dff, timestamps


def get_all_cells_dff(dff_traces_df):
    """gets the dff traces for all cells
    in a BehaviorOphysExperiment

    Parameters
    ----------
    dff_traces_df :pandas dataframe
        dff_traces attribute of a BehaviorOphysExperiment
        object.

    Returns
    -------
    array
        array of arrays with each second level array containing
        the dff timeseries values for a single cell
    """
    dff_trace_array = np.vstack(dff_traces_df['dff'].values)
    return dff_trace_array


def get_cell_dff(dff_traces_df, cell_specimen_id):
    """_summary_

    Parameters
    ----------
    dff_traces_df : pandas dataframe
        dff_traces attribute of a BehaviorOphysExperiment
        object.
    cell_specimen_id : int
       unified id of segmented cell across experiments
       (assigned after cell matching).

    Returns
    -------
    array
       dff timeseries values for the given specified cell
    """

    cell_dff = dff_traces_df.loc[dff_traces_df["cell_specimen_id"] == cell_specimen_id, "dff"].values[0]  # noqa: E501
    return cell_dff


def get_all_cells_mean_dff(dff_traces_df):
    """gets the mean dff trace for all cells
    for a given BehaviorOphysExperiment

    Parameters
    ----------
    dff_traces_df : pandas dataframe
        dff_traces attribute of a BehaviorOphysExperiment
        object.
    Returns
    -------
    array
        mean dff timeseries values for all cells
    """
    mean_dff = gen_utils.average_df_timeseries_values(dff_traces_df, 'dff')
    return mean_dff


def get_transparent_segmentation_mask(ophysObject):
    """transforms the segmentation mask image to
    make the background transparent

    Parameters
    ----------
    ophysObject : (BehaviorOphysExperiment)
        Object provided via allensdk.brain_observatory
        module
    Returns
    -------
    2D image
        1 where an ROI mask and NaN everywhere else
    """
    segmentation_mask = ophysObject.segmentation_mask_image
    transparent_mask = np.zeros(segmentation_mask[0].shape)
    transparent_mask[:] = np.nan
    transparent_mask[segmentation_mask[0] == 1] = 1
    return transparent_mask






from brain_observastory_utilities.datasets.behavior.data_formatting import calculate_response_matrix,\
    calculate_dprime_matrix
import os
import path
import sys
import numpy as np
import matplotlib.pyplot as plt

from .plot_utils import *

sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
import brain_observatory_utilities.datasets.behavior.data_access as data  # noqa: E402


def plot_running(dataObject, ax=None):
    """ plot running speed for trial on specified dataset
    x axis is time (sec), y is running speed (cm/sec)

    Parameters
    ----------
        dataObject : (BehaviorSesson, BehaviorOphysExperiment)
            Objects provided via allensdk.brain_observatory
            module
        ax : (matplotlib.axes), optional
            Figure axes
    Returns
    ----------
    None
    """
    if ax is None:
        fig, ax = plt.subplots()
        ax.set_xlabel("time (sec)")
    speed, timestamps = data.get_running_speed_timeseries(dataObject)
    ax.plot(timestamps, speed,
            color=DATASTREAM_STYLE_DICT['running_speed']['color'])  # noqa: F405, E501
    ax.set_title("running speed")
    ax.set_ylabel("speed (cm/s)")
    return ax


def plot_licks(dataObject, ax=None):
    """ plot licks as tick marks on specified dataset.
    x axis is time (sec) and y axis is lick ticks

    Parameters
    ----------
        dataObject : (BehaviorSesson, BehaviorOphysExperiment)
            Objects provided via allensdk.brain_observatory module
        ax : (matplotlib.axes), optional
            Figure axes
    Returns
    ----------
    None
    """
    if ax is None:
        fig, ax = plt.subplots()

    licks = data.get_lick_timestamps(dataObject)
    ax.plot(licks, np.zeros_like(licks), marker="|",
            linestyle="none",
            color=DATASTREAM_STYLE_DICT['licks']['color'])  # noqa: F405
    return ax


def plot_rewards(dataObject, ax=None, reward_type="all"):
    """ plot rewards as blue dots on specified axis
    x axis is time (sec) and y axis is rewards.

    Parameters
    ----------
        dataObject : (BehaviorSesson, BehaviorOphysExperiment)
            Objects provided via allensdk.brain_observatory module
        ax : (matplotlib.axes), optional
            Figure axes
        reward_type : string
            options:
                "all": all rewards (auto and earned)
                "auto": only free or unearned rewards
                "earned": only earned (hit on a go trial) rewards
    Returns
    ----------
    None
    """
    if ax is None:
        fig, ax = plt.subplots()

    reward_timesestamps = data.get_reward_timestamps(dataObject,
                                                     reward_type=reward_type)
    ax.plot(
        reward_timesestamps,
        np.zeros_like(reward_timesestamps),
        marker="o",
        linestyle="none",
        color='b',
        markersize=6,
        alpha=0.25,
    )
    return ax


def plot_stimuli(dataObject, ax=None):
    """ plot stimuli as colored bars on specified axis

    Parameters
    ----------
        dataObject : (BehaviorSesson, BehaviorOphysExperiment)
            Objects provided via allensdk.brain_observatory
            module
        ax : (matplotlib.axes), optional
            Figure axes
    Returns
    ----------
    None
    """
    if ax is None:
        fig, ax = plt.subplots()

    for _, stimulus in dataObject.stimulus_presentations.iterrows():
        ax.axvspan(stimulus["start_time"],
                   stimulus["stop_time"], alpha=0.5)
    return ax


def plot_pupil_area(dataObject, ax=None):
    """ plot pupil area timeseries.
    x axis is time (sec) and y axis is pupil area
    in pixels squared.

    Parameters
    ----------
        dataObject : (BehaviorOphysExperiment)
            Object provided via allensdk.brain_observatory
            module
        ax : (matplotlib.axes), optional
            Figure axes
    Returns
    ----------
    matplotlib.axes
    """
    if ax is None:
        fig, ax = plt.subplots()
        ax.set_xlabel("time (sec)")

    pupil_area, timestamps = data.get_pupil_area_timeseries(dataObject)  # noqa: E501

    ax.plot(
        timestamps, pupil_area,
        color=DATASTREAM_STYLE_DICT['pupil_area']['color'],  # noqa: F405
    )
    ax.set_title(DATASTREAM_STYLE_DICT['pupil_area']['label'])  # noqa: F405
    ax.set_ylabel("pupil area\n$(pixels^2)$")
    return ax


def plot_response_matrix(stimulus_presentations, ax=None, vmin=0, vmax=1, cmap='viridis', cbar_ax=None):
    '''
    makes a plot of the response matrix given a table of stimuli

    Parameters:
    -----------
    stimulus_presentations : pandas.dataframe
        From experiment.stimulus_presentations, after annotating as follows:
            annotate_stimuli(experiment, inplace = True)
    ax : matplotlib axis
        axis on which to plot.
        If not passed, will create a figure and axis and return both
    vmin : float
        minimum for plot, default = 0
    vmax : float
        maximum for plot, default = 1
    cmap : string
        cmap for plot, default = 'viridis'
    cbar_ax : matplotlib axis
        axis on which to plot colorbar
        colorbar will not be added if not passed

    Returns:
    --------
    fig, ax if ax is not passed
    None if ax is passed
    '''

    return_fig_ax = False
    if ax is None:
        return_fig_ax = True
        fig, ax = plt.subplots()

    response_matrix = calculate_response_matrix(stimulus_presentations)

    im = ax.imshow(
        response_matrix,
        cmap=cmap,
        aspect='equal',
        vmin=vmin,
        vmax=vmax,
    )
    ax.set_xticks(range(len(response_matrix)))
    ax.set_xticklabels(response_matrix.columns)
    ax.set_yticks(range(len(response_matrix)))
    ax.set_yticklabels(response_matrix.index)
    ax.set_xlabel('image name')
    ax.set_ylabel('previous image name')

    if cbar_ax is not None:
        plt.colorbar(im, cax=cbar_ax, label='response probability')

    if return_fig_ax:
        return fig, ax


def plot_dprime_matrix(stimulus_presentations, ax=None, vmin=0, vmax=1.5, cmap='magma', cbar_ax=None):
    '''
    makes a plot of the response matrix given a table of stimuli
    Parameters:
    -----------
    stimulus_presentations : pandas.dataframe
        From experiment.stimulus_presentations, after annotating as follows:
            annotate_stimulus_presentations(experiment, inplace = True)
    ax : matplotlib axis
        axis on which to plot.
        If not passed, will create a figure and axis and return both
    vmin : float
        minimum for plot, default = 0
    vmax : float
        maximum for plot, default = 1.5
    cmap : string
        cmap for plot, default = 'magma'
    cbar_ax : matplotlib axis
        axis on which to plot colorbar
        colorbar will not be added if not passed

    Returns:
    --------
    fig, ax if ax is not passed
    None if ax is passed
    '''

    return_fig_ax = False
    if ax is None:
        return_fig_ax = True
        fig, ax = plt.subplots()

    dprime_matrix = calculate_dprime_matrix(stimulus_presentations)

    im = ax.imshow(
        dprime_matrix,
        cmap=cmap,
        aspect='equal',
        vmin=vmin,
        vmax=vmax,
    )
    ax.set_xticks(range(len(dprime_matrix)))
    ax.set_xticklabels(dprime_matrix.columns)
    ax.set_yticks(range(len(dprime_matrix)))
    ax.set_yticklabels(dprime_matrix.index)
    ax.set_xlabel('image name')
    ax.set_ylabel('previous image name')

    if cbar_ax is not None:
        plt.colorbar(im, cax=cbar_ax, label="d'")

    if return_fig_ax:
        return fig, ax


if __name__ == "__main__":
    # make_plots(experiment_dataset) - test
    pass



from setuptools import setup, find_packages

setup(
    name="brain_observatory_utilities",
    version="0.5.1",
    packages=find_packages(exclude=["tests"]),
    include_package_data=True,
    description="Utilities for analyzing, manipulating and visualizing data for Brain Observatory projects",
    dowload_url='https://github.com/AllenInstitute/brain_observatory_utilities/archive/refs/tags/v0.4.0.tar.gz',
    url="https://github.com/AllenInstitute/mindscope_utilities",
    author="Allen Institute",
    author_email="marinag@alleninstitute.org, corbettb@alleninstitute.org, michaelbu@alleninstitute.org, clark.roll@alleninstitute.org",
    license="Allen Institute",
    install_requires=[
        "flake8",
        "pytest",
        "allensdk",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "License :: Other/Proprietary License",  # Allen Institute Software License
        "Natural Language :: English",
        "Programming Language :: Python :: 3.8",
    ],
)



import numpy as np
import pandas as pd
import numpy as np
from brain_observatory_utilities.utilities.general_utilities import get_trace_average
import brain_observatory_utilities.datasets.behavior.data_access as data_access
from brain_observatory_utilities.utilities import general_utilities

import brain_observatory_utilities.utilities.general_utilities as utilities
import brain_observatory_utilities.datasets.stimulus_alignment.data_formatting as stimulus_alignment


def limit_stimulus_presentations_to_change_detection(stimulus_presentations):
    '''
    if column 'stimulus_block_name' is in stimulus_presentations table (as in SDK v2.16.2),
    limit stimulus presentations table to the change detection block
    '''
    if 'stimulus_block_name' in stimulus_presentations: # This is true for VBO
        stimulus_presentations = stimulus_presentations[stimulus_presentations.stimulus_block_name.str.contains('change_detection')]
        # change a few columns from type Boolean to bool (they were previously Boolean so they could contain NaNs for non-change detection stim blocks)
        # stimulus_presentations = convert_boolean_cols_to_bool(stimulus_presentations)
    else: # For VBO, limit to active behavior block 
        stimulus_presentations = stimulus_presentations[(stimulus_presentations.stimulus_block==0)]

    return stimulus_presentations


def convert_boolean_cols_to_bool(stimulus_presentations):
    '''
    For any dataframe containing columns derived from the stimulus_presentations table,
    go through all columns and identify those that are type boolean (which occurs when the column has NaNs and bools)
    and convert NaNs to False then set dtype to bool.

    This is needed because many operations fail on columns of type boolean.
    Some columns in stimulus_presentations are boolean in new SDK outputs because of the new stimulus_blocks,
    as many values specific to change_detection task are set to NaN in other stimulus blocks, which
    means that the entire column gets the dtype boolean instead of bool.
    '''
    for column in stimulus_presentations.columns.values:
        try:
            if type(stimulus_presentations[column].dtype).__name__ == 'BooleanDtype':
                row_ids = stimulus_presentations[stimulus_presentations[column].isnull()].index
                stimulus_presentations.loc[row_ids, column] = False
                stimulus_presentations[column] = stimulus_presentations[column].astype('bool')
        except:
            if stimulus_presentations[column].dtype == 'boolean':
                # remove NaNs and make bool
                row_ids = stimulus_presentations[stimulus_presentations[column].isnull()].index
                stimulus_presentations.loc[row_ids, column] = False
                stimulus_presentations[column] = stimulus_presentations[column].astype('bool')
    return stimulus_presentations


def add_mean_running_speed_to_stimulus_presentations(stimulus_presentations,
                                                     running_speed,
                                                     time_window=[0, 0.75]):
    '''
    Append a column to stimulus_presentations which contains
    the mean running speed in a range relative to
    the stimulus start time.

    Args:
        stimulus_presentations (pd.DataFrame): dataframe of
            stimulus presentations.
            Must contain: 'start_time'
        running_speed (pd.DataFrame): dataframe of running speed.
            Must contain: 'speed', 'timestamps'
        time_window: array
            timestamps in seconds, relative to the start of each stimulus
            to average the running speed.
            default = [-3,3]
    Returns:
        stimulus_presentations with new column
        "mean_running_speed" containing the
        mean running speed within the specified window
        following each stimulus presentation.

    Example:
        # get visual behavior cache
        from allensdk.brain_observatory.behavior.behavior_project_cache import VisualBehaviorOphysProjectCache as bpc  # noqa E501
        cache_dir = SOME_LOCAL_DIR
        cache = bpc.from_s3_cache(cache_dir=cache_dir)

        # load data for one experiment
        dataset = cache.get_behavior_ophys_experiment(experiment_id)

        # get necessary tables
        stimulus_presentations = dataset.stimulus_presentations.copy()
        running_speed = dataset.running_speed.copy()

        # add running_speed to stim presentations
        stimulus_presentations = add_mean_running_speed_to_stimulus_presentations(stimulus_presentations, running_speed)  # noqa E501
    '''

    stim_running_speed = stimulus_presentations.apply(
        lambda row: utilities.get_trace_average(
            running_speed['speed'].values,
            running_speed['timestamps'].values,
            row["start_time"] + time_window[0],
            row["start_time"] + time_window[1]), axis=1,)
    stimulus_presentations["mean_running_speed"] = stim_running_speed
    return stimulus_presentations


def add_mean_pupil_to_stimulus_presentations(stimulus_presentations, eye_tracking, column_to_use='pupil_area', time_window=[0, 0.75]):
    '''
    Append a column to stimulus_presentations which contains
    the mean pupil area, diameter, or radius in a range relative to
    the stimulus start time.

    Args:
        stimulus_presentations (pd.DataFrame): dataframe of stimulus presentations.  # noqa E501
            Must contain: 'start_time'
        eye_tracking (pd.DataFrame): dataframe of eye tracking data.
            Must contain: timestamps', 'pupil_area', 'pupil_width', 'likely_blinks'
        time_window (list with 2 elements): start and end of the range
            relative to the start of each stimulus to average the pupil area.
        column_to_use: column in eyetracking table to use to get mean, options: 'pupil_area', 'pupil_width', 'pupil_radius', 'pupil_diameter'
                        if 'pupil_diameter' or 'pupil_radius' are provided, they will be calculated from 'pupil_area'
                        if 'pupil_width' is provided, the column 'pupil_width' will be directly used from eye_tracking table
    Returns:
        stimulus_presentations table with new column "mean_pupil_"+column_to_use with the
        mean pupil value within the specified window following each stimulus presentation.

    Example:
        # get visual behavior cache
        from allensdk.brain_observatory.behavior.behavior_project_cache import VisualBehaviorOphysProjectCache as bpc  # noqa E501
        cache_dir = r'\\allen\programs\braintv\workgroups\nc-ophys\visual_behavior\platform_paper_cache'
        cache = bpc.from_s3_cache(cache_dir=cache_dir)

        # load data for one experiment
        dataset = cache.get_behavior_ophys_experiment(experiment_id)

        # get necessary tables
        stimulus_presentations = dataset.stimulus_presentations.copy()
        eye_tracking = dataset.eye_tracking.copy()

        # add pupil area to stim presentations
        stimulus_presentations = add_mean_pupil_to_stimulus_presentations(stimulus_presentations, eye_tracking, column_to_use='pupil_area')  # noqa E501
    '''

    eye_tracking = data_access.get_pupil_data(eye_tracking, interpolate_likely_blinks=True, normalize_to_gray_screen=True, zscore=False,
                                              interpolate_to_ophys=False, ophys_timestamps=None, stimulus_presentations=stimulus_presentations)

    eye_tracking_timeseries = eye_tracking[column_to_use].values
    mean_pupil_around_stimulus = stimulus_presentations.apply(
        lambda row: utilities.get_trace_average(
            eye_tracking_timeseries,
            eye_tracking['timestamps'].values,
            row["start_time"] + time_window[0],
            row["start_time"] + time_window[1],
        ), axis=1,)
    stimulus_presentations["mean_"+column_to_use] = mean_pupil_around_stimulus
    return stimulus_presentations


def add_rewards_to_stimulus_presentations(stimulus_presentations,
                                          rewards,
                                          time_window=[0, 3]):
    '''
    Append a column to stimulus_presentations which contains
    the timestamps of rewards that occur
    in a range relative to the onset of the stimulus.

    Args:
        stimulus_presentations (pd.DataFrame): dataframe of
            stimulus presentations.
            Must contain: 'start_time'
        rewards (pd.DataFrame): rewards dataframe. Must contain 'timestamps'
        time_window (list with 2 elements): start and end of the range
            relative to the start of each stimulus
            to average the running speed.
    Returns:
        stimulus_presentations with a new column called "reward" that contains
        reward times that fell within the window relative to each stim time

    Example:
        # get visual behavior cache
        from allensdk.brain_observatory.behavior.behavior_project_cache import VisualBehaviorOphysProjectCache as bpc  # noqa E501
        cache_dir = SOME_LOCAL_DIRECTORY
        cache = bpc.from_s3_cache(cache_dir=cache_dir)

        # load data for one experiment
        dataset = cache.get_behavior_ophys_experiment(experiment_id)

        # get necessary tables
        stimulus_presentations = dataset.stimulus_presentations.copy()
        rewards = dataset.rewards.copy()

        # add rewards to stim presentations
        stimulus_presentations = add_rewards_to_stimulus_presentations(stimulus_presentations, rewards)  # noqa E501
    '''

    reward_times = rewards['timestamps'].values
    rewards_each_stim = stimulus_presentations.apply(
        lambda row: reward_times[
            ((reward_times > row["start_time"] + time_window[0]) & (reward_times < row["start_time"] + time_window[1]))],
        axis=1,
    )
    stimulus_presentations["rewards"] = rewards_each_stim
    return stimulus_presentations


def add_licks_to_stimulus_presentations(stimulus_presentations,
                                        licks,
                                        time_window=[0, 0.75]):
    '''
    Append a column to stimulus_presentations which
    contains the timestamps of licks that occur
    in a range relative to the onset of the stimulus.

    Args:
        stimulus_presentations (pd.DataFrame): 
            dataframe of stimulus presentations.
            Must contain: 'start_time'
        licks (pd.DataFrame): lick dataframe. Must contain 'timestamps'
        time_window (list with 2 elements): start and end of the range
            relative to the start of each stimulus to average the running speed.  # noqa E501
    Returns:
        stimulus_presentations with a new column called "licks" that contains
        lick times that fell within the window relative to each stim time


    Example:
        # get visual behavior cache
        from allensdk.brain_observatory.behavior.behavior_project_cache import VisualBehaviorOphysProjectCache as bpc  # noqa E501
        cache_dir = SOME_LOCAL_DIRECTORY
        cache = bpc.from_s3_cache(cache_dir=cache_dir)

        # load data for one experiment
        dataset = cache.get_behavior_ophys_experiment(experiment_id)

        # get necessary tables
        stimulus_presentations = dataset.stimulus_presentations.copy()
        licks = dataset.licks.copy()

        # add licks to stim presentations
        stimulus_presentations = add_licks_to_stimulus_presentations(stimulus_presentations, licks)
    '''

    lick_times = licks['timestamps'].values
    licks_each_stim = stimulus_presentations.apply(
        lambda row: lick_times[((lick_times > row["start_time"] + time_window[0]) & (lick_times < row["start_time"] + time_window[1]))],
        axis=1)
    stimulus_presentations["licks"] = licks_each_stim

    # Add first lick time and lick latency relative to stim onset
    first_lick_times = []
    for i in range(len(stimulus_presentations)):
        if len(stimulus_presentations.iloc[i]['licks'])>0: 
            first_lick_times.append(stimulus_presentations.iloc[i]['licks'][0])
        else: 
            first_lick_times.append(np.nan)
    stimulus_presentations['first_lick_time'] = first_lick_times
    stimulus_presentations['lick_latency'] = stimulus_presentations['first_lick_time'] - stimulus_presentations['start_time']
    
    return stimulus_presentations


def calculate_reward_rate(response_latency=None,
                          starttime=None,
                          window=0.75,
                          trial_window=25,
                          initial_trials=10):

    assert len(response_latency) == len(starttime)

    df = pd.DataFrame({'response_latency': response_latency,
                       'starttime': starttime})

    # adds a column called reward_rate to the input dataframe
    # the reward_rate column contains a rolling average of rewards/min
    # window sets the window in which a response is considered correct,
    # so a window of 1.0 means licks before 1.0 second are considered correct
    #
    # Reorganized into this unit-testable form by Nick Cain April 25 2019

    reward_rate = np.zeros(len(df))
    # make the initial reward rate infinite,
    # so that you include the first trials automatically.
    reward_rate[:initial_trials] = np.inf

    for trial_number in range(initial_trials, len(df)):

        min_index = np.max((0, trial_number - trial_window))
        max_index = np.min((trial_number + trial_window, len(df)))
        df_roll = df.iloc[min_index:max_index]

        # get a rolling number of correct trials
        correct = len(df_roll[df_roll.response_latency < window])

        # get the time elapsed over the trials
        time_elapsed = df_roll.starttime.iloc[-1] - df_roll.starttime.iloc[0]

        # calculate the reward rate, rewards/min
        reward_rate_on_this_lap = correct / time_elapsed * 60

        reward_rate[trial_number] = reward_rate_on_this_lap
    return reward_rate


def add_reward_rate_to_stimulus_presentations(stimulus_presentations, trials):
    '''
    Parameters:
    ____________
    trials: Pandas.DataFrame
        dataset.trials
    stimulus_presentation: Pandas.DataFrame
        ophys_datasetexperiment.stimulus_presentations

    Returns:
    ___________
    stimulus_presentation: Pandas.DataFrame
        with 'reward_rate_trials' column

    'reward_rate' is calculated by the SDK based on the rolling reward rate over trials (not stimulus presentations)
    https://github.com/AllenInstitute/AllenSDK/blob/master/allensdk/brain_observatory/behavior/trials_processing.py#L941
    '''

    last_time = 0
    reward_rate_by_frame = []

    # need to calculate response latency because SDK doesnt include it for VBN
    if 'change_time_no_display_delay' in trials.keys(): # this means its from VBN
        trials['response_latency'] = trials.response_time-trials.change_time_no_display_delay
        trials['change_time'] = trials.change_time_no_display_delay

    # recalculate reward_rate for trials 
    trials['reward_rate'] = calculate_reward_rate(trials['response_latency'].values, trials['start_time'])

    trials = trials[trials['aborted'] == False]  # NOQA
    for change_time in trials.change_time.values:
        reward_rate = trials[trials.change_time ==  # NOQA
                             change_time].reward_rate.values[0]
        # add reward rate value from trial to all stim presentations belonging to that trial
        for start_time in stimulus_presentations.start_time: 
            if (start_time < change_time) and (start_time > last_time):
                reward_rate_by_frame.append(reward_rate)
                last_time = start_time
    # fill the last flashes with last value
    for i in range(len(stimulus_presentations) - len(reward_rate_by_frame)):
        reward_rate_by_frame.append(reward_rate_by_frame[-1])

    stimulus_presentations['reward_rate'] = reward_rate_by_frame
    return stimulus_presentations


def add_engagement_state_to_stimulus_presentations(
        stimulus_presentations, trials):
    """
    Add 'engaged' Boolean column and 'engagement_state' string ('engaged' or 'disengaged')
    using threshold of  2 rewards per minute, with reward_rate calculated as in the SDK by the
    function add_reward_rate_to_stimulus_presentations() in this repo, which is a copy of what is done in the SDK.
    Previously this function pulled directly from the SDK, but the funciton was added to a class and is no longer directly accessible.

    :param stimulus_presentations: stimulus_presentations attribute of SDK dataset object
    :param trials: trials attribute of SDK dataset object
    :return: stimulus_presentations with columns added: 'reward_rate', 'engaged', 'engagement_state'
    """

    if 'reward_rate' not in stimulus_presentations.keys():
        stimulus_presentations = add_reward_rate_to_stimulus_presentations(stimulus_presentations, trials)
    
    reward_threshold = 2

    stimulus_presentations['engaged'] = [x > reward_threshold for x in stimulus_presentations['reward_rate'].values]
    stimulus_presentations['engagement_state'] = ['engaged' if engaged else 'disengaged' for engaged in stimulus_presentations['engaged'].values]

    return stimulus_presentations


def add_epochs_to_stimulus_presentations(stimulus_presentations, time_column='start_time', epoch_duration_mins=10):
    """
    Add column called 'epoch' with values as an index for the epoch within a session, for a given epoch duration.

    :param stimulus_presentations: dataframe with a column indicating event start times
    :param time_column: name of column in dataframe indicating event times
    :param epoch_duration_mins: desired epoch length in minutes
    :return: input dataframe with epoch column added
    """
    start_time = stimulus_presentations[time_column].values[0]
    stop_time = stimulus_presentations[time_column].values[-1]
    epoch_times = np.arange(start_time, stop_time, epoch_duration_mins * 60)
    stimulus_presentations['epoch'] = None
    for i, time in enumerate(epoch_times):
        if i < len(epoch_times) - 1:
            indices = stimulus_presentations[(stimulus_presentations[time_column] >= epoch_times[i]) &
                                             (stimulus_presentations[time_column] < epoch_times[i + 1])].index.values
        else:
            indices = stimulus_presentations[(
                stimulus_presentations[time_column] >= epoch_times[i])].index.values
        stimulus_presentations.loc[indices, 'epoch'] = i
    return stimulus_presentations


def add_trials_id_to_stimulus_presentations(stimulus_presentations, trials):
    """
    Add trials_id to stimulus presentations by finding the closest change time to each stimulus start time
    If there is no corresponding change time, the trials_id is NaN
    :param: stimulus_presentations: stimulus_presentations attribute of SDK dataset object, must have 'start_time'
    :param trials: trials attribute of SDK dataset object, must have 'change_time'
    """
    # make sure the trials table has a `change_time` column (its called `change_time_no_display_lag` for VBN)
    if 'change_time' not in trials.columns: 
        trials['change_time'] = trials['change_time_no_display_delay'] 

    # for each stimulus_presentation, find the trials_id that is closest to the start time
    # add to a new column called 'trials_id'
    for idx, stimulus_presentation in stimulus_presentations.iterrows():
        start_time = stimulus_presentation['start_time']
        query_string = 'start_time > @start_time - 0.5 and start_time < @start_time + 0.5'
        trials_id = (np.abs(start_time - trials.query(query_string)['change_time']))
        if len(trials_id) == 1:
            trials_id = trials_id.idxmin()
        else:
            trials_id = np.nan
        stimulus_presentations.loc[idx, 'trials_id'] = trials_id
    return stimulus_presentations


# def add_change_trials_id_to_stimulus_presentations(stimulus_presentations, trials):
#     """
#     Add trials_id for change and sham change times to stimulus presentations by
#     finding the closest change (or sham change) time to each stimulus start time.
#     Column is called 'change_trials_id' to distinguish from 'trials_id' which is applied to all flashes in a trial
#     If there is no corresponding change (or sham change) time, the change_trials_id is NaN
#     i.e. this function only assigns a change_trials_id to the stimulus presentations corresponding to go and catch trials as defined in the trials table.
#     :param: stimulus_presentations: stimulus_presentations attribute of SDK dataset object, must have 'start_time'
#     :param trials: trials attribute of SDK dataset object, must have 'change_time'
#     """
#     # for each stimulus_presentation, find the change_time that is closest to the start time
#     # then add the corresponding trials_id to stimulus_presentations
#     trials = trials.copy()
#     if 'change_time' not in trials.keys():
#         trials['change_time'] = trials['change_time_no_display_delay']
    
#     stimulus_presentations = stimulus_presentations.copy()
#     for row in range(len(stimulus_presentations)):
#         this_start_time = stimulus_presentations.iloc[row].start_time
#         # if its not the last row / stimulus presentation
#         if row <= len(stimulus_presentations)-2:
#             # get the start time of the next stimulus
#             next_start_time = stimulus_presentations.iloc[row+1].start_time
#         else:
#             # if it is the last row, infer that the next start would be 750ms from now
#             next_start_time = stimulus_presentations.iloc[row].start_time + 0.75
#         # find the trial where change time falls between the current and next stimulus start times
#         trial_data = trials[(trials.change_time>this_start_time) & (trials.change_time<=next_start_time)]
#         if len(trial_data) > 0:
#             trials_id = trial_data.index.values[0]
#         else:
#             trials_id = np.nan
#         stimulus_presentations.loc[row, 'change_trials_id'] = trials_id
#     return stimulus_presentations


def add_change_trials_id_to_stimulus_presentations(stimulus_presentations, trials):
    """
    Add trials_id to stimulus presentations by finding the closest change time to each stimulus start time
    If there is no corresponding change time, the trials_id is NaN
    :param: stimulus_presentations: stimulus_presentations attribute of SDK dataset object, must have 'start_time'
    :param trials: trials attribute of SDK dataset object, must have 'change_time'
    """
    # make sure the trials table has a `change_time` column (its called `change_time_no_display_lag` for VBN)
    if 'change_time' not in trials.columns: 
        trials['change_time'] = trials['change_time_no_display_delay'] 

    # for each stimulus_presentation, find the trials_id that is closest to the start time
    # add to a new column called 'trials_id'
    for idx, stimulus_presentation in stimulus_presentations.iterrows():
        start_time = stimulus_presentation['start_time']
        query_string = 'change_time > @start_time - 0.5 and change_time < @start_time + 0.5'
        trials_id = (np.abs(start_time - trials.query(query_string)['change_time']))
        if len(trials_id) == 1:
            trials_id = trials_id.idxmin()
        else:
            trials_id = np.nan
        stimulus_presentations.loc[idx, 'change_trials_id'] = trials_id
    return stimulus_presentations


def add_trial_type_to_stimulus_presentations(stimulus_presentations, trials):
    """
    Add trials_id to stimulus presentations table
    then join trial type columns of trials table with stimulus_presentations
    adds trial type value to all rows of stimulus_presentations belonging to a trial
    
    trial types = ['aborted', 'auto_rewarded', 'go', 'catch']
    :param: stimulus_presentations: stimulus_presentations attribute of SDK dataset object, must have 'start_time'
    :param trials: trials attribute of SDK dataset object, must have 'start_time'
    """
    # make sure the trials table has a `change_time` column (its called `change_time_no_display_lag` for VBN)
    if 'change_time' not in trials.columns: 
        trials['change_time'] = trials['change_time_no_display_delay'] 

    # add trials_id for all stimulus presentations and merge to get trial type information
    # if 'change_trials_id' not in stimulus_presentations.columns:
    #     stimulus_presentations = add_change_trials_id_to_stimulus_presentations(stimulus_presentations, trials)
    # get trial type columns
    trials = trials[['go', 'catch', 'aborted', 'auto_rewarded']]
    # merge trial type columns into stimulus_presentations
    stimulus_presentations = stimulus_presentations.reset_index().merge(trials, on='trials_id', how='left')
    stimulus_presentations = stimulus_presentations.set_index('stimulus_presentations_id')
    return stimulus_presentations


def add_trials_data_to_stimulus_presentations_table(stimulus_presentations, trials):
    """
    Add trials_id to stimulus presentations table then join relevant columns of trials with stimulus_presentations
    :param: stimulus_presentations: stimulus_presentations attribute of SDK dataset object, must have 'start_time'
    :param trials: trials attribute of SDK dataset object, must have 'change_time'
    """
    
    # make sure the trials table has a `change_time` column (its called `change_time_no_display_lag` for VBN)
    if 'change_time' not in trials.columns: 
        trials['change_time'] = trials['change_time_no_display_delay'] 

    # add trials_id and merge to get trial type information
    if 'change_trials_id' not in stimulus_presentations.columns:
        stimulus_presentations = add_change_trials_id_to_stimulus_presentations(stimulus_presentations, trials)
    # only keep certain columns
    columns_to_keep = ['change_time', 'go', 'catch', 'auto_rewarded', #'aborted' - there is no change time on aborts to merge on
                     'hit', 'miss', 'false_alarm', 'correct_reject',
                     'response_time', 'reward_time', 'reward_volume']
    # if 'response_latency' in trials.columns: # VBN doesnt have response_latency
    #     columns_to_keep = columns_to_keep + ['response_latency']
    trials = trials[columns_to_keep]
    # Add change_trials_id column to trials to merge with stim presentations
    trials['change_trials_id'] = trials.index.values
    # merge trials columns into stimulus_presentations, how='left' to ensure only change trials are merged
    stimulus_presentations = stimulus_presentations.reset_index().merge(trials, on='change_trials_id', how='left')
    stimulus_presentations = stimulus_presentations.set_index('stimulus_presentations_id')
    
    return stimulus_presentations


def add_change_trial_outcomes_to_stimulus_presentations(stimulus_presentations, trials):
    """
    Add trials_id to stimulus presentations table, just for go and catch trials,
    then join columns of trials table indicating whether a go trial was hit or miss
    and whether a catch trial was a false alarm or correct reject
    with stimulus_presentations,
    relevant columns here are
    :param: stimulus_presentations: stimulus_presentations attribute of SDK dataset object, must have 'start_time'
    :param trials: trials attribute of SDK dataset object, must have 'change_time'
    """
    # add trials_id and merge to get trial type information
    stimulus_presentations = add_change_trials_id_to_stimulus_presentations(stimulus_presentations, trials)
    # only keep certain columns
    trials = trials[['change_time', 'hit', 'miss',
                     'false_alarm', 'correct_reject',
                     'reward_time', 'reward_volume']]
    # merge trials columns into stimulus_presentations

    stimulus_presentations = stimulus_presentations.reset_index().merge(trials, left_on='change_trials_id',
                                                                        right_on='trials_id', how='left')
    stimulus_presentations = stimulus_presentations.set_index('stimulus_presentations_id')
    return stimulus_presentations


def add_time_from_last_change_to_stimulus_presentations(stimulus_presentations):
    '''
    Adds a column to stimulus_presentations called 'time_from_last_change', which is the time, in seconds since the last image change

    ARGS: SDK session object
    MODIFIES: session.stimulus_presentations
    RETURNS: stimulus_presentations
    '''
    stimulus_times = stimulus_presentations["start_time"].values
    change_times = stimulus_presentations[stimulus_presentations.is_change]['start_time'].values
    time_from_last_change = general_utilities.time_from_last(stimulus_times, change_times)
    stimulus_presentations["time_from_last_change"] = time_from_last_change

    return stimulus_presentations


def add_time_from_last_omission_to_stimulus_presentations(stimulus_presentations):
    '''
    Adds a column to stimulus_presentations called 'time_from_last_omission',
    which is the time, in seconds since the last stimulus omission

    ARGS: SDK stimulus presentations table
    MODIFIES: session.stimulus_presentations
    RETURNS: stimulus_presentations
    '''
    stimulus_times = stimulus_presentations["start_time"].values
    omission_times = stimulus_presentations[stimulus_presentations['omitted']==True]['start_time'].values
    time_from_last_omission = general_utilities.time_from_last(stimulus_times, omission_times)
    stimulus_presentations["time_from_last_omitted"] = time_from_last_omission

    return stimulus_presentations


def add_time_from_last_lick_to_stimulus_presentations(stimulus_presentations, licks):
    '''
    Adds a column to stimulus_presentations called 'time_from_last_lick',
    which is the time, in seconds since the last lick response

    ARGS: SDK stimulus presentations table, SDK licks table
    MODIFIES: session.stimulus_presentations
    RETURNS: stimulus_presentations
    '''

    stimulus_presentations = stimulus_presentations.copy()

    stimulus_times = stimulus_presentations["start_time"].values
    lick_times = licks.timestamps.values
    time_from_last_lick = general_utilities.time_from_last(stimulus_times, lick_times)
    stimulus_presentations["time_from_last_lick"] = time_from_last_lick

    return stimulus_presentations


def add_time_from_last_reward_to_stimulus_presentations(stimulus_presentations, rewards):
    '''
    Adds a column to stimulus_presentations called 'time_from_last_reward',
    which is the time, in seconds since the last reward was delivered

    ARGS: SDK stimulus presentations table, SDK rewards table
    MODIFIES: session.stimulus_presentations
    RETURNS: stimulus_presentations
    '''
    stimulus_times = stimulus_presentations["start_time"].values
    reward_times = rewards.timestamps.values
    time_from_last_reward = general_utilities.time_from_last(stimulus_times, reward_times)
    stimulus_presentations["time_from_last_reward"] = time_from_last_reward

    return stimulus_presentations


def add_time_from_last_to_stimulus_presentations(stimulus_presentations, licks, rewards):
    """
    Adds multiple columns to stimulus presentations with the number of stimuli that have elapsed
    since the last change, omission, lick, and reward using the Boolean columns 'is_change', 'omitted', 'licked', and 'rewarded'
    If these Boolean columns are not included in stimulus_presentations, they will be added

    :param stimulus_presentations: SDK stimulus_presentations table
    :return: stimulus_presentations with 'stimulus_count_from_last' columns added

    """
    stimulus_presentations = stimulus_presentations.copy()
    stimulus_presentations = add_time_from_last_change_to_stimulus_presentations(stimulus_presentations)
    stimulus_presentations = add_time_from_last_omission_to_stimulus_presentations(stimulus_presentations)
    stimulus_presentations = add_time_from_last_reward_to_stimulus_presentations(stimulus_presentations, rewards)
    stimulus_presentations = add_time_from_last_lick_to_stimulus_presentations(stimulus_presentations, licks)

    return stimulus_presentations


def stimulus_count_from_last(stimulus_presentations, column_to_count_from='is_change'):
    """
    Takes a Boolean column in stimulus_presentations and counts the number of stimulus presentations
    after each True instance of the column_to_count_from column value,
    adds a column called 'stimulus_count_from_last_'+column_to_count_from that contains that count value

    Example: if using column_to_count_from = 'is_change', and the first 10 stimuli are [F, F, F, F, T, F, F, F, F],
    the [stimulus_count_since_last_is_change' column will be [1, 2, 3, 4, 0, 1, 2, 3, 4]

    :param stimulus_presentations: SDK stimulus presentations table, must include Boolean column for column_to_count_from
    :param column_to_count_from: column value in stimulus_presentations to use to count number of stimulus presentations
                                since the column value was last True
    :return: stimulus_presentations with added column 'stimulus_count_from_last_'+column_to_count_from
    """
    stimulus_presentations = stimulus_presentations.copy()
    # label each stimulus presentation based on the number of stimuli since the trial start
    stimulus_presentations['stimulus_count_from_last_' + column_to_count_from] = None
    stimulus_count = 0
    # iterate over every stimulus
    for idx, row in stimulus_presentations.iterrows():
        current_value = row[column_to_count_from]
        # if current_value is not True, increment the counter
        if current_value == False:
            stimulus_count += 1
        elif current_value == True:  # if current value is True, reset the counter
            stimulus_count = 0
        else:
            stimulus_count = np.nan
        stimulus_presentations.at[idx, 'stimulus_count_from_last_' + column_to_count_from] = stimulus_count

    return stimulus_presentations


def add_stimulus_count_from_last_to_stimulus_presentations(stimulus_presentations, licks, rewards):
    """
    Adds multiple columns to stimulus presentations with the number of stimuli that have elapsed
    since the last change, omission, lick, and reward using the Boolean columns 'is_change', 'omitted', 'licked', and 'rewarded'
    If these Boolean columns are not included in stimulus_presentations, they will be added

    :param stimulus_presentations: SDK stimulus_presentations table
    :return: stimulus_presentations with 'stimulus_count_from_last' columns added

    """
    stimulus_presentations = stimulus_presentations.copy()
    if 'licked' not in stimulus_presentations.keys():
        stimulus_presentations = add_licks_to_stimulus_presentations(stimulus_presentations, licks,
                                                                              time_window=[0, 0.75])
    if 'rewarded' not in stimulus_presentations.keys():
        stimulus_presentations = add_rewards_to_stimulus_presentations(stimulus_presentations, rewards,
                                                                                time_window=[0, 0.75])
    for column in ['is_change', 'omitted', 'licked', 'rewarded']:
        stimulus_presentations = stimulus_count_from_last(stimulus_presentations, column_to_count_from=column)

    return stimulus_presentations


def add_n_to_stimulus_presentations(stimulus_presentations):
    """
    Adds a column to stimulus_presentations called 'n_after_change',
    which is the number of stimulus presentations that have occurred since the last change.
    It will also add a column called 'n_after_omission',
    which is the number of stimulus presentations that have occurred since the last omission,
    before the next change.
    If there is no omission, this value will be -1.
    Presentations before the first change or omission will have a value of -1.
    It will also add a column called 'n_before_change',
    which is the number of stimulus presentations that have occurred before the next change.
    Presentations after the last change will have a value of -1.
    Presentations before the first change will also have a value of -1.
    0 for 'n_after_change' and 'n_before_change' indicates the change itself.
    0 for 'n_after_omission' indicates the omission itself.

    Parameters
    ----------
    stimulus_presentations : pd.DataFrame
        stimulus_presentations table from SDK dataset object

    Returns
    -------
    stimulus_presentations : pd.DataFrame
        stimulus_presentations table with 'n_after_change', 'n_after_omission', and 'n_before_change' columns added
    """

    change_ind = stimulus_presentations[stimulus_presentations['is_change']].index.values

    # Adding n_after_change
    # -1 indicates before the first change
    n_after_change = np.zeros(len(stimulus_presentations)) - 1
    for i in range(1, len(change_ind)):
        n_after_change[change_ind[i - 1]: change_ind[i]
                       ] = np.arange(0, change_ind[i] - change_ind[i - 1]).astype(int)
    n_after_change[change_ind[i]:] = np.arange(
        0, len(stimulus_presentations) - change_ind[i]).astype(int)
    stimulus_presentations['n_after_change'] = n_after_change

    # Adding n_before_change
    # -1 indicates after the last and before the first change
    n_before_change = np.zeros(len(stimulus_presentations)) - 1
    for i in range(len(change_ind) - 1):
        n_before_change[change_ind[i] + 1: change_ind[i + 1] + 1] = np.arange(
            change_ind[i + 1] - change_ind[i] - 1, -1, -1).astype(int)
    stimulus_presentations['n_before_change'] = n_before_change

    # Adding n_after_omission
    # -1 indicates before the first omission or
    n_after_omission = np.zeros(len(stimulus_presentations)) - 1
    # from the next change till the next omission # noqa E114,E116
    # if there are no omissions, n_after_omission will be all -1
    # and 'omitted' will be added and assigned to False
    if 'omitted' in stimulus_presentations.columns:
        omission_ind = stimulus_presentations[stimulus_presentations['omitted']].index.values
        for i in range(len(omission_ind)):
            if change_ind[-1] > omission_ind[i]:  # if there is a change after the omission
                next_change_ind = change_ind[change_ind > omission_ind[i]][0]
                n_after_omission[omission_ind[i]: next_change_ind] = np.arange(
                    0, next_change_ind - omission_ind[i]).astype(int)
            else:
                n_after_omission[omission_ind[i]:] = np.arange(
                    0, len(stimulus_presentations) - omission_ind[i]).astype(int)
    else:
        stimulus_presentations['omitted'] = False
    stimulus_presentations['n_after_omission'] = n_after_omission

    return stimulus_presentations


def add_stimulus_count_within_trial_to_stimulus_presentations(stimulus_presentations, trials):
    """
    Add a column to stimulus_presentations that indicates how many stimuli have been shown since the trial start

    :param: stimulus_presentations: stimulus_presentations attribute of dataset object, must have 'start_time'
    :param trials: trials attribute of dataset object, must have 'start_time'
    """
    stimulus_presentations = stimulus_presentations.copy()
    # if trials_id is not a column of stimulus_presentations, add it
    if 'trials_id' not in stimulus_presentations.keys():
        print('trials_id does not exist in stimulus_presentations, please add it before running the function add_stimulus_count_within_trial_to_stimulus_presentations' )
        # stimulus_presentations = add_trials_id_to_stimulus_presentations(stimulus_presentations, trials)

    # label each stimulus presentation based on the number of stimuli since the trial start
    stimulus_presentations['stimulus_count_within_trial'] = None
    stimulus_count_within_trial = 0
    last_trial_id = -1 # start at -1 as baseline to compare with first trials_id (which is 0)
    # iterate over every stimulus
    for idx, row in stimulus_presentations.iterrows():
        trials_id = row.trials_id
        # if trials_id is the same as the last trial, increment stimulus_count_within_trial
        if trials_id == last_trial_id:
            stimulus_count_within_trial += 1
        else:
            stimulus_count_within_trial = 0
            # the current trial will be the last_trial_id for the next row / iteration
            last_trial_id = trials_id
        stimulus_presentations.at[idx, 'stimulus_count_within_trial'] = stimulus_count_within_trial

    return stimulus_presentations


def add_could_change_to_stimulus_presentations(stimulus_presentations, trials, licks):
    """
    Adds a column called could_change to the stimulus presentations table that indicates whether a given
    stimulus presentation could have been a change or sham change based on the known change time distribution.
    A stimulus could change if it is at least 4 stimulus flashes after the trial start,
    and there was not a lick on the previous stimulus, and the current or next stimulus is not omitted.
    :param stimulus_presentations: AllenSDK stimulus presentations table
    :param trials: AllenSDK trials table
    :param licks: AllenSDK licks table
    :return:
    """

    licks = licks.copy()
    trials = trials.copy()
    stimulus_presentations = stimulus_presentations.copy()
    # if trials_id is not a column of stimulus_presentations, add it
    # if 'trials_id' not in stimulus_presentations.keys():
    #     stimulus_presentations = add_trials_id_to_stimulus_presentations(stimulus_presentations, trials)
        # if trials_id is not a column of stimulus_presentations, add it
    if 'licks' not in stimulus_presentations.keys():
        stimulus_presentations = add_licks_to_stimulus_presentations(stimulus_presentations, licks)
        # if trials_id is not a column of stimulus_presentations, add it
    if 'licked' not in stimulus_presentations.keys():
        stimulus_presentations['licked'] = None
        stimulus_presentations['licked'] = [True if len(licks) > 0 else False for licks in stimulus_presentations.licks.values]
    if 'stimulus_count_within_trial' not in stimulus_presentations.keys():
        stimulus_presentations = add_stimulus_count_within_trial_to_stimulus_presentations(stimulus_presentations, trials)

    # add previous_image_name
    stimulus_presentations['previous_image_name'] = stimulus_presentations['image_name'].shift()
    # add previous_response_on_trial
    stimulus_presentations['previous_response_on_trial'] = False
    stimulus_presentations['previous_change_on_trial'] = False
    # set 'stimulus_presentations_id' and 'trials_id' as indices to speed lookup
    stimulus_presentations = stimulus_presentations.reset_index().set_index(['stimulus_presentations_id', 'trials_id'])
    for idx, row in stimulus_presentations.iterrows():
        stim_id, trials_id = idx
        # get all stimuli before the current on the current trial
        mask = (stimulus_presentations.index.get_level_values(0) < stim_id) & (
            stimulus_presentations.index.get_level_values(1) == trials_id)
        # check to see if any previous stimuli have a response lick
        stimulus_presentations.at[idx, 'previous_response_on_trial'] = stimulus_presentations[mask]['licked'].any()
        stimulus_presentations.at[idx, 'previous_change_on_trial'] = stimulus_presentations[mask]['is_change'].any()
    # set the index back to being just 'stimulus_presentations_id'
    stimulus_presentations = stimulus_presentations.reset_index().set_index('stimulus_presentations_id')

    # add could_change column to indicate whether the stimulus presentation falls into the range in which changes can occur
    stimulus_presentations['could_change'] = False
    for idx, row in stimulus_presentations.iterrows():
        # check if we meet conditions where a change could occur on this stimulus (at least 4th flash of trial, no previous change on trial)
        # changes can only happen after the 4th stimulus within a trial
        # changes can only happen if there is no previous response on the trial (i.e. not on aborted trials)
        # changes can only happen if there is no previous change on that trial (i.e. not during the grace period / consumption window)
        # changes can only happen if the previous stimulus was not omitted
        if row['stimulus_count_within_trial'] >= 4 and row['previous_response_on_trial'] is False \
                and row['previous_change_on_trial'] is False \
                and row['image_name'] != 'omitted' \
                and row['previous_image_name'] != 'omitted':
            stimulus_presentations.at[idx, 'could_change'] = True

    return stimulus_presentations


def annotate_stimuli(dataset, inplace=False):
    '''
    adds the following columns to the stimulus_presentations table, facilitating calculation
    of behavior performance based entirely on the stimulus_presentations table:

    'trials_id': the corresponding ID of the trial in the trials table in which the stimulus occurred
    'previous_image_name': the name of the stimulus on the last flash (will list 'omitted' if last stimulus is omitted)
    'next_start_time': The time of the next stimulus start (including the time of the omitted stimulus if the next stimulus is omitted)
    'auto_rewarded': True for trials where rewards were delivered regardless of animal response
    'trial_stimulus_index': index of the given stimulus on the current trial. For example, the first stimulus in a trial has index 0, the second stimulus in a trial has index 1, etc
    'response_lick': Boolean, True if a lick followed the stimulus
    'response_lick_times': list of all lick times following this stimulus
    'response_lick_latency': time difference between first lick and stimulus
    'previous_response_on_trial': Boolean, True if there has been a lick to a previous stimulus on this trial
    'could_change': Boolean, True if the stimulus met the conditions that would have allowed
                    to be chosen as the change stimulus by camstim:
                        * at least the fourth stimulus flash in the trial
                        * not preceded by any licks on that trial

    Parameters:
    -----------
    dataset : an SDK session object
        AllenSDK BehaviorOphysExperiment object
        or AllenSDK BehaviorEcephysSession object
        or AllenSDK BehaviorSession object
        See:
        https://github.com/AllenInstitute/AllenSDK/blob/master/allensdk/brain_observatory/behavior/behavior_ophys_experiment.py  # noqa E501
        https://github.com/AllenInstitute/AllenSDK/blob/master/allensdk/brain_observatory/ecephys/behavior_ecephys_session.py  # noqa E501
    
    inplace : Boolean
        If True, operates on the dataset.stimulus_presentations object directly and returns None
        If False (default), operates on a copy and returns the copy

    Returns:
    --------
    Pandas.DataFrame (if inplace == False)
    None (if inplace == True)
    '''

    if inplace:
        stimulus_presentations = dataset.stimulus_presentations
    else:
        stimulus_presentations = dataset.stimulus_presentations.copy()

    trials = dataset.trials.copy()
    if 'change_time' not in trials.keys():
        trials['change_time'] = trials['change_time_no_display_delay']
    if 'stimulus_count_within_trial' not in stimulus_presentations.keys():
        stimulus_presentations = add_stimulus_count_within_trial_to_stimulus_presentations(stimulus_presentations, trials)
    
    # limit to change detection block
    stimulus_presentations = limit_stimulus_presentations_to_change_detection(stimulus_presentations)

    # add previous_image_name
    stimulus_presentations['previous_image_name'] = stimulus_presentations['image_name'].shift()

    # add next_start_time
    stimulus_presentations['next_start_time'] = stimulus_presentations['start_time'].shift(-1)

    # add trials_id and trial_stimulus_index
    # stimulus_presentations['trials_id'] = None
    stimulus_presentations['trial_stimulus_index'] = None
    last_trial_id = -1
    trial_stimulus_index = 0

    # add response_lick, response_lick_times, response_lick_latency
    stimulus_presentations['response_lick'] = False
    stimulus_presentations['response_lick_times'] = None
    stimulus_presentations['response_lick_latency'] = None

    # make a copy of trials with 'start_time' as index to speed lookup
    trials = trials.copy().reset_index().set_index('start_time')
    
    # make a copy of licks with 'timestamps' as index to speed lookup
    licks = dataset.licks.copy().reset_index().set_index('timestamps')

    # iterate over every stimulus
    for idx, row in stimulus_presentations.iterrows():
        # trials_id is last trials_id with start_time <= stimulus_time
        try:
            trials_id = trials.loc[:row['start_time']].iloc[-1]['trials_id']
        except IndexError:
            trials_id = -1
        # stimulus_presentations.at[idx, 'trials_id'] = trials_id

        if trials_id == last_trial_id:
            trial_stimulus_index += 1
        else:
            trial_stimulus_index = 0
            last_trial_id = trials_id
        stimulus_presentations.at[idx, 'trial_stimulus_index'] = trial_stimulus_index

        # note the `- 1e-9` acts as a <, as opposed to a <=
        stim_licks = licks.loc[row['start_time']:row['next_start_time'] - 1e-9].index.to_list()

        stimulus_presentations.at[idx, 'response_lick_times'] = stim_licks
        if len(stim_licks) > 0:
            stimulus_presentations.at[idx, 'response_lick'] = True
            stimulus_presentations.at[idx, 'response_lick_latency'] = stim_licks[0] - row['start_time']

    # merge in auto_rewarded column from trials table
    stimulus_presentations = stimulus_presentations.reset_index().merge(
        trials[['auto_rewarded', 'trials_id']],
        on='trials_id',
        how='left',
    ).set_index('stimulus_presentations_id')

    # Add could change
    stimulus_presentations = add_could_change_to_stimulus_presentations(stimulus_presentations, trials, dataset.licks)
    
    if inplace is False:
        return stimulus_presentations
    

def add_behavior_info_to_stimulus_presentations(stimulus_presentations, trials, licks, rewards,
                                                running_speed, eye_tracking):
    """
    Adds a variety of useful columns to the stimulus presentations table by incorporating information
    from the trials, licks, rewards, running_speed and eye_tracking tables.
    Useful to filter stimuli based on behavioral information.
    Added columns include trials_id, go, catch, aborted, auto-rewarded, hit, miss, false_alarm, correct_reject,
    as well as lick and reward times for each stimulus presentation,
    reward rate in rewards / min (computed across trials not stimulus presentations),
    and mean running speed and pupil width for each stimulus

    :param stimulus_presentations: SDK stimulus_presentations table
    :param trials: SDK trials table
    :param licks: SDK licks table
    :param rewards: SDK rewards table
    :param running_speed: SDK running_speed table
    :param eye_tracking: SDK eye tracking table, will use pupil_width column of eye tracking table
    :return:
    """

    trials = trials.copy()
    if 'change_time' not in trials.keys():
        trials['change_time'] = trials['change_time_no_display_delay']
    stimulus_presentations = stimulus_presentations.copy()
    # add columns from trials table: ['change_time', 'hit', 'miss', 'false_alarm', 'correct_reject', 'reward_time', 'reward_volume']
    # applied only to stimulus presentations corresponding to change and sham change times (go and catch)
    stimulus_presentations = add_change_trials_id_to_stimulus_presentations(stimulus_presentations, trials)
    stimulus_presentations = add_change_trial_outcomes_to_stimulus_presentations(stimulus_presentations, trials)
    # add columns from trials table: ['go', 'catch', 'aborted', 'auto_rewarded']
    # applied to all stimulus presentations belonging to that trial
    # stimulus_presentations = add_trials_id_to_stimulus_presentations(stimulus_presentations, trials)
    # stimulus_presentations = add_trial_type_to_stimulus_presentations(stimulus_presentations, trials)
    # add the timing of licks and rewards for each stimulus presentation
    stimulus_presentations = add_licks_to_stimulus_presentations(stimulus_presentations, licks, time_window=[0, 0.75])
    stimulus_presentations = add_rewards_to_stimulus_presentations(stimulus_presentations, rewards, time_window=[0, 0.75])
    # add reward rate across trials to stimulus presentations
    stimulus_presentations = add_reward_rate_to_stimulus_presentations(stimulus_presentations, trials)
    # add mean running speed and pupil diameter for each stimulus presentation
    stimulus_presentations = add_mean_running_speed_to_stimulus_presentations(stimulus_presentations, running_speed, time_window=[0, 0.75])
    stimulus_presentations = add_mean_pupil_to_stimulus_presentations(stimulus_presentations, eye_tracking,
                                             column_to_use='pupil_width', time_window=[0, 0.75])

    return stimulus_presentations


def add_timing_info_to_stimulus_presentations(stimulus_presentations, trials, licks, rewards):
    """
    Annotate stimulus presentations table with information about timing relative to events of interest,
    such as time from last change / omission / lick / reward,
    whether a stimulus was before or after a change or an omission,
    and whether a given stimulus could have been a change based on known change time distribution (4-12 flashes from the time of the last lick)

    :param stimulus_presentations: SDK stimulus presentations table
    :param trials: SDK trials table
    :param licks: SDK licks table
    :param rewards: SDK rewards table
    :return:
    """

    trials = trials.copy()
    if 'change_time' not in trials.keys():
        trials['change_time'] = trials['change_time_no_display_delay']
        
    # add columns indicating wither previous or subsequent stimulus presentations are a change or omission
    stimulus_presentations['pre_change'] = stimulus_presentations['is_change'].shift(-1)
    stimulus_presentations['post_change'] = stimulus_presentations['is_change'].shift(1)
    stimulus_presentations['pre_omitted'] = stimulus_presentations['omitted'].shift(-1)
    stimulus_presentations['post_omitted'] = stimulus_presentations['omitted'].shift(1)

    # add stimulus count within each behavioral trial
    stimulus_presentations = add_stimulus_count_within_trial_to_stimulus_presentations(stimulus_presentations, trials)
    # add stimulus count from last change / omission / lick / reward
    stimulus_presentations = add_stimulus_count_from_last_to_stimulus_presentations(stimulus_presentations, licks, rewards)
    # add time elapsed since last change / omission / lick / reward
    stimulus_presentations = add_time_from_last_to_stimulus_presentations(stimulus_presentations, licks, rewards)

    # add column indicating whether a given stimulus presentations could have been a change stimulus
    # given the known change time distribution and whether or not the mouse licked on a given stimulus (thus aborting and reseting the trial)
    stimulus_presentations = add_could_change_to_stimulus_presentations(stimulus_presentations, trials, licks)

    return stimulus_presentations


def get_annotated_stimulus_presentations(dataset, epoch_duration_mins=10):
    """
    Takes in an SDK dataset object and returns the stimulus_presentations table with additional columns.
    Adds several useful columns to the stimulus_presentations table, including the mean running speed and pupil diameter for each stimulus,
    the times of licks for each stimulus, the rolling reward rate, an identifier for 10 minute epochs within a session,
    whether or not a stimulus was a pre-change or pre or post omission, and whether change stimuli were hits or misses
    :param dataset: obj
        AllenSDK BehaviorOphysExperiment object
        or AllenSDK BehaviorEcephysSession object
        or AllenSDK BehaviorSession object
        See:
        https://github.com/AllenInstitute/AllenSDK/blob/master/allensdk/brain_observatory/behavior/behavior_ophys_experiment.py  # noqa E501
        https://github.com/AllenInstitute/AllenSDK/blob/master/allensdk/brain_observatory/ecephys/behavior_ecephys_session.py  # noqa E501
    
    :return: stimulus_presentations attribute of dataset object, with additional columns added
    """
    stimulus_presentations = dataset.stimulus_presentations.copy()
    # limit to change detection block
    stimulus_presentations = limit_stimulus_presentations_to_change_detection(stimulus_presentations)

    trials = dataset.trials.copy()
    if 'change_time' not in trials.keys(): 
        trials['change_time'] = trials['change_time_no_display_delay']

    # add licks
    stimulus_presentations = add_licks_to_stimulus_presentations(
        stimulus_presentations, dataset.licks, time_window=[0, 0.75])
    # add running
    stimulus_presentations = add_mean_running_speed_to_stimulus_presentations(
        stimulus_presentations, dataset.running_speed, time_window=[0, 0.75])
    # if hasattr('ophys_experiment', 'eye_tracking'):
    try:
        stimulus_presentations = add_mean_pupil_to_stimulus_presentations(
            stimulus_presentations,
            dataset.eye_tracking,
            column_to_use='pupil_width',
            time_window=[0, 0.75])
        
    except Exception as e:
        print('could not add mean pupil to stimulus presentations, length of eye_tracking attribute is', len(
                dataset.eye_tracking))
        print(e)

    # add trials info
    # try:  # not all session types have catch trials or omissions
    stimulus_presentations = add_trials_data_to_stimulus_presentations_table(
        stimulus_presentations, trials)
    # add time from last change
    stimulus_presentations = add_time_from_last_change_to_stimulus_presentations(stimulus_presentations)
    # # add whether a flash could have been a change based on the change time distribution

    stimulus_presentations = add_could_change_to_stimulus_presentations(stimulus_presentations, trials, dataset.licks)
    # add pre-change
    stimulus_presentations['pre_change'] = stimulus_presentations['is_change'].shift(-1)
    # add licked Boolean
    if 'licked' not in stimulus_presentations.keys():
        stimulus_presentations['licked'] = [True if len(licks) > 0 else False for licks in stimulus_presentations.licks.values]
    stimulus_presentations['lick_on_next_flash'] = stimulus_presentations['licked'].shift(-1)
    # add omission annotation
    stimulus_presentations['pre_omitted'] = stimulus_presentations['omitted'].shift(-1)
    stimulus_presentations['post_omitted'] = stimulus_presentations['omitted'].shift(1)
    # # add repeat number
    # except Exception as e:
    #     print(e)

    # add reward rate
    stimulus_presentations = add_reward_rate_to_stimulus_presentations(
        stimulus_presentations, trials)
    # add engagement state based on reward rate 
    stimulus_presentations = add_engagement_state_to_stimulus_presentations(
            stimulus_presentations, trials)
    # add epochs
    stimulus_presentations = add_epochs_to_stimulus_presentations(
        stimulus_presentations,
        time_column='start_time',
        epoch_duration_mins=epoch_duration_mins)

    return stimulus_presentations



def calculate_response_matrix(stimuli, aggfunc=np.mean, sort_by_column=True, engaged_only=True):
    '''
    calculates the response matrix for each individual image pair in the `stimulus` dataframe

    Parameters:
    -----------
    stimuli: Pandas.DataFrame
        From experiment.stimulus_presentations, after annotating as follows:
            annotate_stimuli(experiment, inplace = True)
    aggfunc: function
        function to apply to calculation. Default = np.mean
        other options include np.size (to get counts) or np.median
    sort_by_column: Boolean
        if True (default), sorts outputs by column means
    engaged_only: Boolean
        If True (default), calculates only on engaged trials
        Will throw an assertion error if True and 'engagement_state' column does not exist

    Returns:
    --------
    Pandas.DataFrame
        matrix of response probabilities for each image combination
        index = previous image
        column = current image
        catch trials are on diagonal

    '''
    stimuli_to_analyze = stimuli.query(
        'auto_rewarded == False and could_change == True and image_name != "omitted" and previous_image_name != "omitted"')
    if engaged_only:
        assert 'engagement_state' in stimuli_to_analyze.columns, 'stimuli must have column called "engagement_state" if passing engaged_only = True'
        stimuli_to_analyze = stimuli_to_analyze.query(
            'engagement_state == "engaged"')

    response_matrix = pd.pivot_table(
        stimuli_to_analyze,
        values='response_lick',
        index=['previous_image_name'],
        columns=['image_name'],
        aggfunc=aggfunc
    ).astype(float)

    if sort_by_column:
        sort_by = response_matrix.mean(axis=0).sort_values().index
        response_matrix = response_matrix.loc[sort_by][sort_by]

    response_matrix.index.name = 'previous_image_name'

    return response_matrix


def get_licks_df(dataset):
    '''
    Creates a dataframe containing columns for 'timestamps', 'licks', where values are from
    a binary array of the length of stimulus timestamps where frames with no lick are 0 and frames with a lick are 1,
    and a column called 'lick_rate' with values of 'licks' averaged over a 6 frame window to get licks per 100ms,
    Can be used to plot stim triggered average lick rate
    Parameters:
    -----------
    dataset: obj
        AllenSDK BehaviorOphysExperiment object
        or AllenSDK BehaviorEcephysSession object
        or AllenSDK BehaviorSession object
        See:
        https://github.com/AllenInstitute/AllenSDK/blob/master/allensdk/brain_observatory/behavior/behavior_ophys_experiment.py  # noqa E501
        https://github.com/AllenInstitute/AllenSDK/blob/master/allensdk/brain_observatory/ecephys/behavior_ecephys_session.py  # noqa E501
    
    Returns:
    --------
    Pandas.DataFrame with columns 'timestamps', 'licks', and 'lick_rate' in units of licks / 100ms

    '''
    timestamps = dataset.stimulus_timestamps.copy()
    licks = dataset.licks.copy()
    lick_array = np.zeros(timestamps.shape)
    lick_array[licks.frame.values] = 1
    licks_df = pd.DataFrame(data=timestamps, columns=['timestamps'])
    licks_df['licks'] = lick_array
    licks_df['lick_rate'] = licks_df['licks'].rolling(
        window=6, min_periods=1, win_type='triang').mean()

    return licks_df


def calculate_dprime_matrix(stimuli, sort_by_column=True, engaged_only=True):
    '''
    calculates the d' matrix for each individual image pair in the `stimulus` dataframe

    Parameters:
    -----------
    stimuli: Pandas.DataFrame
        From experiment.stimulus_presentations, after annotating as follows:
            annotate_stimuli(experiment, inplace = True)
    sort_by_column: Boolean
        if True (default), sorts outputs by column means
    engaged_only: Boolean
        If True (default), calculates only on engaged trials
        Will throw an assertion error if True and 'engagement_state' column does not exist

    Returns:
    --------
    Pandas.DataFrame
        matrix of d' for each image combination
        index = previous image
        column = current image
        catch trials are on diagonal

    '''
    if engaged_only:
        assert 'engagement_state' in stimuli.columns, 'stimuli must have column called "engagement_state" if passing engaged_only = True'

    response_matrix = calculate_response_matrix(
        stimuli,
        aggfunc=np.mean,
        sort_by_column=sort_by_column,
        engaged_only=engaged_only)

    d_prime_matrix = response_matrix.copy()
    for row in response_matrix.columns:
        for col in response_matrix.columns:
            d_prime_matrix.loc[row][col] = mindscope_utilities.dprime(
                hit_rate=response_matrix.loc[row][col],
                fa_rate=response_matrix[col][col],
                limits=False
            )
            if row == col:
                d_prime_matrix.loc[row][col] = np.nan

    return d_prime_matrix









def create_lick_rate_df(dataset, bin_size=6):
    '''
    Returns a dataframe containing columns for 'timestamps', 'licks', and 'lick_rate'. where values are from
    'licks' is a binary array of the length of stimulus timestamps where frames with no lick are 0 and frames with a lick are 1,
    'lick_rate' contains values of 'licks' averaged over a rolling window using the provided 'bin_size' in units of acquisition frames
    Can be used to compute event triggered average lick rate using brain_observatory_utilities.utilities.general_utilities.event_triggered_response

    Parameters:
    -----------
    dataset: obj
        AllenSDK BehaviorOphysExperiment object, BehaviorSession object, or BehaviorEcephysSession object
        See:
        https://github.com/AllenInstitute/AllenSDK/blob/master/allensdk/brain_observatory/behavior/behavior_ophys_experiment.py  # noqa E501
        https://github.com/AllenInstitute/AllenSDK/blob/master/allensdk/brain_observatory/behavior/behavior_session.py  # noqa E501
        https://github.com/AllenInstitute/AllenSDK/blob/master/allensdk/brain_observatory/ecephys/behavior_ecephys_session.py  # noqa E501
    bin_size: int
        number of frames (timestamps) to average over to get rolling lick rate
        default = 6 frames to give lick rate in licks / 100ms (assuming 60Hz acquisition rate for licks)
    Returns:
    --------
    Pandas.DataFrame with columns 'timestamps', 'licks', and 'lick_rate' in units of licks / 100ms
    '''

    timestamps = dataset.stimulus_timestamps.copy()
    licks = dataset.licks.copy()
    lick_array = np.zeros(timestamps.shape)
    lick_array[licks.frame.values] = 1
    licks_df = pd.DataFrame(data=timestamps, columns=['timestamps'])
    licks_df['licks'] = lick_array
    licks_df['lick_rate'] = licks_df['licks'].rolling(window=bin_size, min_periods=1, win_type='triang').mean()

    return licks_df


def filter_eye_tracking(eye_tracking, interpolate_likely_blinks=False, normalize_to_gray_screen=False, zscore=False,
                   interpolate_to_ophys=False, ophys_timestamps=None, stimulus_presentations=None):
    """
    removes 'likely_blinks' from all columns in eye_tracking and converts pupil_area to pupil_diameter and pupil_radius
    interpolates over NaNs resulting from removing likely_blinks if interpolate = true
    :param eye_tracking: eye_tracking attribute of AllenSDK BehaviorOphysExperiment object
    :param column_to_use: 'pupil_area', 'pupil_width', 'pupil_diameter', or 'pupil_radius'
                            'pupil_area' and 'pupil_width' are existing columns of the eye_tracking table
                            'pupil_diameter' and 'pupil_radius' are computed from 'pupil_area' column
    :param interpolate: Boolean, whether or not to interpolate points where likely_blinks occured

    :return:
    """
    import scipy

    # set index to timestamps so they dont get overwritten by subsequent operations
    eye_tracking = eye_tracking.set_index('timestamps')

    # compute pupil_diameter and pupil_radius from pupil_area
    eye_tracking['pupil_diameter'] = np.sqrt(eye_tracking.pupil_area) / np.pi  # convert pupil area to pupil diameter
    eye_tracking['pupil_radius'] = np.sqrt(eye_tracking['pupil_area'] * (1 / np.pi))  # convert pupil area to pupil radius

    # set all timepoints that are likely blinks to NaN for all eye_tracking columns
    if True in eye_tracking.likely_blink.unique(): # only can do this if there are likely blinks to filter out
        eye_tracking.loc[eye_tracking['likely_blink'], :] = np.nan

    # add timestamps column back in
    eye_tracking['timestamps'] = eye_tracking.index.values

    # interpolate over likely blinks, which are now NaNs
    if interpolate_likely_blinks:
        eye_tracking = eye_tracking.interpolate()

    # divide all columns by average of gray screen period prior to behavior session
    if normalize_to_gray_screen:
        assert stimulus_presentations is not None, 'must provide stimulus_presentations if normalize_to_gray_screen is True'
        spontaneous_frames = stimulus_alignment.get_spontaneous_frames(stimulus_presentations,
                                                             eye_tracking.timestamps.values,
                                                             gray_screen_period_to_use='before')
        for column in eye_tracking.keys():
            if (column != 'timestamps') and (column!='likely_blink'):
                gray_screen_mean_value = np.nanmean(eye_tracking[column].values[spontaneous_frames])
                eye_tracking[column] = eye_tracking[column] / gray_screen_mean_value
    # z-score pupil data
    if zscore:
        for column in eye_tracking.keys():
            if (column != 'timestamps') and (column != 'likely_blink'):
                eye_tracking[column] = scipy.stats.zscore(eye_tracking[column], nan_policy='omit')

    # interpolate to ophys timestamps
    if interpolate_to_ophys:
        assert ophys_timestamps is not None, 'must provide ophys_timestamps if interpolate_to_ophys is True'
        eye_tracking_ophys_time = pd.DataFrame({'timestamps': ophys_timestamps})
        for column in eye_tracking.keys():
            if (column != 'timestamps') and (column != 'likely_blink'):
                f = scipy.interpolate.interp1d(eye_tracking['timestamps'], eye_tracking[column], bounds_error=False)
                eye_tracking_ophys_time[column] = f(eye_tracking_ophys_time['timestamps'])
                eye_tracking_ophys_time[column].fillna(method='ffill', inplace=True)
        eye_tracking = eye_tracking_ophys_time

    return eye_tracking



from allensdk.core.reference_space_cache import ReferenceSpaceCache
import re
import nrrd
import os
import requests

class VBN_CCF:

    def __init__(self, manifest_path='manifest.json', resolution=10):
        reference_space_key = 'annotation/ccf_2017'
        self.resolution = resolution
        self.rspc = ReferenceSpaceCache(resolution, reference_space_key, manifest=manifest_path)
        # ID 1 is the adult mouse structure graph
        self.tree = self.rspc.get_structure_tree(structure_graph_id=1) 
        self._annotation = None
        self._streamlines = None
        self._manifest_directory = os.path.dirname(manifest_path)


    @property
    def annotation(self):
        if self._annotation is None:
            self._annotation, meta = self.rspc.get_annotation_volume()
        return self._annotation

    
    @property
    def streamlines(self):
        if self._streamlines is None:
            streamlines_path = os.path.join(self._manifest_directory, 'laplacian_10.nrrd')
            #First check to see if the streamlines have already been downloaded
            if os.path.exists(streamlines_path):
                self._streamlines, header = nrrd.read(streamlines_path)
            
            #Otherwise download it and write it to file
            else:
                s = requests.get('https://www.dropbox.com/sh/7me5sdmyt5wcxwu/AACFY9PQ6c79AiTsP8naYZUoa/laplacian_10.nrrd?dl=1')
                with open(streamlines_path, 'wb') as f:
                    f.write(bytes(s.content))
                self._streamlines, header = nrrd.read(streamlines_path)
                
        return self._streamlines


    def get_structure_by_acronym(self, acronym):
        try:
            structure = self.tree.get_structures_by_acronym([acronym])
        except KeyError:
            print(f'Could not find structure corresponding to acronym {acronym}')
            structure = [{}]
        return structure


    def get_structure_name_by_acronym(self, acronym):
        structure = self.get_structure_by_acronym(acronym)[0]
        return structure.get('name', None)

    
    def get_structure_id_by_coordinate(self, ap_coord, dv_coord, lr_coord):

        volume_coords = [int(coord/self.resolution) for coord in [ap_coord, dv_coord, lr_coord]]
        shape = self.annotation.shape
        if any([(v<0 or v>=s) for v,s in zip(volume_coords,shape)]):
            print(f'Coordinate {[ap_coord, dv_coord, lr_coord]} is outside ccf')
            id = 0
        else:
            id = self.annotation[volume_coords[0], volume_coords[1], volume_coords[2]]
        return id


    def get_structure_by_id(self, id):
        if id == 0:
            structure = [{'name': 'outside_brain',
                        'acronym': 'outside_brain'}]
        else:
            structure = self.tree.get_structures_by_id([id])
        return structure


    def get_structure_by_coordinate(self, ap_coord, dv_coord, lr_coord):
        try:
            id = self.get_structure_id_by_coordinate(ap_coord, dv_coord, lr_coord)
            structure = self.get_structure_by_id(id)
        except Exception as e:
            print(f'Could not get structure corresponding to id: {id} due to error {e}')
            structure = [{}]
        return structure


    def get_structure_acronym_by_coordinate(self, ap_coord, dv_coord, lr_coord):

        structure = self.get_structure_by_coordinate(ap_coord, dv_coord, lr_coord)[0]
        return structure.get('acronym', None)

    
    def get_layer_name(self, acronym):

        if acronym in ['CA1', 'CA2', 'CA3']:
            return ''
    
        try:
            first_num = re.findall(r'\d+', acronym)[0]
            first_num_ind = acronym.find(first_num)
            if first_num_ind<0:
                return ''
            
            layer = acronym[first_num_ind:]
            return layer

        except IndexError:
            return ''
    

    def get_cortical_depth_by_coordinate(self, ap_coord, dv_coord, lr_coord):
        volume_coords = [int(coord/10) for coord in [ap_coord, dv_coord, lr_coord]]
        try:
            cortical_depth = self.streamlines[volume_coords[0], volume_coords[1], volume_coords[2]]
        except IndexError as e:
            print(f'Coordinate {[ap_coord, dv_coord, lr_coord]} is outside of CCF')
            cortical_depth = 0
        return cortical_depth


import numpy as np
import pandas as pd
from tqdm import tqdm
import xarray

from brain_observatory_utilities.utilities import general_utilities
from brain_observatory_utilities.datasets.behavior import data_formatting as behavior


def build_tidy_cell_df(ophys_experiment, exclude_invalid_rois=True):
    '''
    Builds a tidy dataframe describing activity for every cell in ophys_experiment.
    Tidy format is defined as one row per observation.
    Thus, the output dataframe will be n_cells x n_timetpoints long

    Parameters:
    -----------
    ophys_experiment : AllenSDK BehaviorOphysExperiment object
        A BehaviorOphysExperiment instance
        See https://github.com/AllenInstitute/AllenSDK/blob/master/allensdk/brain_observatory/behavior/behavior_ophys_ophys_experiment.py  # noqa E501
    exclude_invalid_rois : bool
        If True (default), only includes ROIs that are listed as `valid_roi = True` in the ophys_experiment.cell_specimen_table.
        If False, include all ROIs.
        Note that invalid ROIs are only exposed for internal AllenInstitute users, so passing `False` will not change behavior for external users

    Returns:
    --------
    Pandas.DataFrame
        Tidy Format (one observation per row) with the following columns:
            * timestamps (float) : the ophys timestamps
            * cell_roi_id (int) : the cell roi id
            * cell_specimen_id (int) : the cell specimen id
            * dff (float) : measured deltaF/F for every timestep
            * events (float) : extracted events for every timestep
            * filtered events (float) : filtered (convolved with half-gaussian) events for every timestep
    '''

    # make an empty list to populate with dataframes for each cell
    list_of_cell_dfs = []

    # query on valid_roi if exclude_invalid_rois == True
    if exclude_invalid_rois:
        cell_specimen_table = ophys_experiment.cell_specimen_table.query('valid_roi').reset_index()  # noqa E501
    else:
        cell_specimen_table = ophys_experiment.cell_specimen_table.reset_index()  # noqa E501

    # iterate over each individual cell
    for idx, row in cell_specimen_table.iterrows():
        cell_specimen_id = row['cell_specimen_id']

        # build a tidy dataframe for this cell
        cell_df = pd.DataFrame({
            'timestamps': ophys_experiment.ophys_timestamps,
            'dff': ophys_experiment.dff_traces.loc[cell_specimen_id]['dff'] if cell_specimen_id in ophys_experiment.dff_traces.index else [np.nan] * len(ophys_experiment.ophys_timestamps),  # noqa E501
            'events': ophys_experiment.events.loc[cell_specimen_id]['events'] if cell_specimen_id in ophys_experiment.events.index else [np.nan] * len(ophys_experiment.ophys_timestamps),  # noqa E501
            'filtered_events': ophys_experiment.events.loc[cell_specimen_id]['filtered_events'] if cell_specimen_id in ophys_experiment.events.index else [np.nan] * len(ophys_experiment.ophys_timestamps),  # noqa E501
        })

        # Make the cell_roi_id and cell_specimen_id columns categorical.
        # This will reduce memory useage since the columns
        # consist of many repeated values.
        for cell_id in ['cell_roi_id', 'cell_specimen_id']:
            cell_df[cell_id] = np.int32(row[cell_id])
            cell_df[cell_id] = pd.Categorical(
                cell_df[cell_id],
                categories=cell_specimen_table[cell_id].unique()
            )

        # append the dataframe for this cell to the list of cell dataframes
        list_of_cell_dfs.append(cell_df)

    # concatenate all dataframes in the list
    tidy_df = pd.concat(list_of_cell_dfs)

    # return the tidy dataframe
    return tidy_df


def get_event_timestamps(
        stimulus_presentation,
        event_type='all',
        onset='start_time'):
    '''
    Gets timestamps of events of interest from the stimulus_presentations df.

    Parameters:
    ___________
    stimulus_presentation: Pandas.DataFrame
        Output of stimulus_presentations with stimulus trial metadata
    event_type: str
        Event of interest. Event_type can be any column in the stimulus_presentation,  # noqa E501
        including 'omissions' or 'change'. Default is 'all', gets all trials  # noqa E501
    onset: str
        optons: 'start_time' - onset of the stimulus, 'stop_time' - offset of the stimulus
        stimulus_presentationshas a multiple timestamps to align data to. Default = 'start_time'.

    Returns:
        event_times: array
        event_ids: array
    --------
    '''
    if event_type == 'all':
        event_times = stimulus_presentation[onset]
        event_ids = stimulus_presentation.index.values
    elif event_type == 'images':
        event_times = stimulus_presentation[stimulus_presentation['omitted'] == False][onset]  # noqa E501
        event_ids = stimulus_presentation[stimulus_presentation['omitted'] == False].index.values  # noqa E501
    elif event_type == 'omissions' or event_type == 'omitted':
        event_times = stimulus_presentation[stimulus_presentation['omitted']][onset]  # noqa E501
        event_ids = stimulus_presentation[stimulus_presentation['omitted']].index.values  # noqa E501
    elif event_type == 'changes' or event_type == 'is_change':
        event_times = stimulus_presentation[stimulus_presentation['is_change']][onset]  # noqa E501
        event_ids = stimulus_presentation[stimulus_presentation['is_change']].index.values  # noqa E501
    else:
        event_times = stimulus_presentation[stimulus_presentation[event_type]][onset]  # noqa E501
        event_ids = stimulus_presentation[stimulus_presentation[event_type]].index.values  # noqa E501

    return event_times, event_ids



def get_stimulus_response_xr(ophys_experiment,
                             data_type='dff',
                             event_type='all',
                             time_window=[-3, 3],
                             response_window_duration=0.5,
                             interpolate=True,
                             output_sampling_rate=None,
                             exclude_invalid_rois=True,
                             **kwargs):
    '''
    Parameters:
    ___________
    ophys_experiment: obj
        AllenSDK BehaviorOphysExperiment object
        A BehaviorOphysExperiment instance
        See https://github.com/AllenInstitute/AllenSDK/blob/master/allensdk/brain_observatory/behavior/behavior_ophys_ophys_experiment.py  # noqa E501
    data_type: str
        neural or behavioral data type to extract, options are: dff (default), events, filtered_events, running_speed, pupil_diameter, lick_rate
    event_type: str
        event type to align to, which can be found in columns of ophys_experiment.stimulus_presentations df.
        options are: 'all' (default) - gets all stimulus trials
                     'images' - gets only image presentations (changes and not changes)
                     'omissions' - gets only trials with omitted stimuli
                     'changes' - get only trials of image changes
    time_window: array
        array of two int or floats indicating the time window on sliced data, default = [-3, 3]
    response_window_duration: float
        time period, in seconds, relative to stimulus onset to compute the mean and baseline response
    interpolate: bool
        type of alignment. If True (default) - interpolates neural data to align timestamps
        with stimulus presentations. If False - shifts data to the nearest timestamps
    output_sampling_rate : float
        Desired sampling of output.
        Input data will be interpolated to this sampling rate if interpolate = True (default). # NOQA E501
        If passing interpolate = False, the sampling rate of the input timeseries will # NOQA E501
        be used and output_sampling_rate should not be specified.
    exclude_invalid_rois : bool
        If True, only ROIs deemed as 'valid' by the classifier will be returned. If False, 'invalid' ROIs will be returned
        This only works if the provided dataset was loaded using internal Allen Institute database, does not work for NWB files.
        In the case that dataset object is loaded through publicly released NWB files, only 'valid' ROIs will be returned


    kwargs: key, value mappings
        Other keyword arguments are passed down to general_utilities.event_triggered_response(),
        for interpolation method such as include_endpoint.

    Returns:
    __________
    stimulus_response_xr: xarray
        Xarray of aligned neural data with multiple dimentions: cell_specimen_id,
        'eventlocked_timestamps', and 'stimulus_presentations_id'

    '''

    import brain_observatory_utilities.datasets.behavior.data_access as behavior_data_access

    # load stimulus_presentations table
    stimulus_presentations = ophys_experiment.stimulus_presentations.copy()

    # limit to change detection block
    stimulus_presentations = behavior.limit_stimulus_presentations_to_change_detection(stimulus_presentations)

    # get event times and event ids (original order in the stimulus flow)
    event_times, event_ids = get_event_timestamps(stimulus_presentations, event_type)

    if ('running' in data_type) or ('pupil' in data_type) or ('lick' in data_type):
        # for behavioral datastreams
        # set up variables to handle only one timeseries per event instead of multiple cell_specimen_ids
        unique_id_string = 'trace_id'  # create a column to take the place of 'cell_specimen_id'
        unique_ids = [0]  # list to iterate over
    else:
        unique_id_string = 'cell_specimen_id'

    if 'running' in data_type:
        data = ophys_experiment.running_speed.copy() # running_speed attribute is already in tidy format
        data = data.rename(columns={'speed':'running_speed'}) # rename column so its consistent with data_type
        data[unique_id_string] = 0  # only one value because only one trace
    elif 'pupil' in data_type:
        data = ophys_experiment.eye_tracking.copy() # eye tracking attribute is in tidy format
        data = behavior_data_access.get_pupil_data(data, interpolate_likely_blinks=True, normalize_to_gray_screen=True, zscore=False,
                                interpolate_to_ophys=False, stimulus_presentations=stimulus_presentations, ophys_timestamps=None)
        # normalize to gray screen baseline
        data[unique_id_string] = 0  # only one value because only one trace
    elif 'lick' in data_type:
        data = behavior.get_licks_df(ophys_experiment) # create dataframe with info about licks for each stimulus timestamp
        data[unique_id_string] = 0  # only one value because only one trace
    else:
        # load neural data
        data = build_tidy_cell_df(ophys_experiment, exclude_invalid_rois=exclude_invalid_rois)
        # all cell specimen ids in an ophys_experiment
        unique_ids = np.unique(data['cell_specimen_id'].values)

    # get native sampling rate if one is not provided
    if output_sampling_rate is None:
        output_sampling_rate = 1 / np.diff(data['timestamps']).mean()

    # collect aligned data
    sliced_dataout = []

    # align data using interpolation method
    for unique_id in tqdm(unique_ids):
        etr = general_utilities.event_triggered_response(
            data=data[data[unique_id_string] == unique_id],
            t='timestamps',
            y=data_type,
            event_times=event_times,
            t_start=time_window[0],
            t_end=time_window[1],
            output_format='wide',
            interpolate=interpolate,
            output_sampling_rate=output_sampling_rate,
            **kwargs
        )

        # get timestamps array
        trace_timebase = etr.index.values

        # collect aligned data from all cell, all trials into one array
        sliced_dataout.append(etr.transpose().values)

    # convert to xarray
    sliced_dataout = np.array(sliced_dataout)
    stimulus_response_xr = xarray.DataArray(
        data=sliced_dataout,
        dims=(unique_id_string, 'stimulus_presentations_id',
              'eventlocked_timestamps'),
        coords={
            'eventlocked_timestamps': trace_timebase,
            'stimulus_presentations_id': event_ids,
            unique_id_string: unique_ids
        }
    )


    # get traces for significance computation
    if 'events' in data_type:
        traces_array = np.vstack(ophys_experiment.events[data_type].values)
    elif data_type == 'dff':
        traces_array = np.vstack(ophys_experiment.dff_traces['dff'].values)
    else:
        traces_array = data[data_type].values

    # compute mean activity following stimulus onset and during pre-stimulus baseline
    stimulus_response_xr = compute_means_xr(stimulus_response_xr, response_window_duration=response_window_duration)

    # get mean response for each trial
    mean_responses = stimulus_response_xr.mean_response.data.T  # input needs to be array of nConditions, nCells

    try:
        # compute significance of each trial, returns array of nConditions, nCells
        p_value_gray_screen = get_p_value_from_shuffled_spontaneous(mean_responses,
                                                                    stimulus_presentations,
                                                                    ophys_experiment.ophys_timestamps,
                                                                    traces_array,
                                                                    response_window_duration*output_sampling_rate,
                                                                    output_sampling_rate)
    except:
        p_value_gray_screen = np.zeros(mean_responses.shape)

    # put p_value_gray_screen back into same coordinates as xarray and make it an xarray data array
    p_value_gray_screen = xarray.DataArray(data=p_value_gray_screen.T, coords=stimulus_response_xr.mean_response.coords)

    # create new xarray with means and p-values
    stimulus_response_xr = xarray.Dataset({
        'eventlocked_traces': stimulus_response_xr.eventlocked_traces,
        'mean_response': stimulus_response_xr.mean_response,
        'mean_baseline': stimulus_response_xr.mean_baseline,
        'p_value_gray_screen': p_value_gray_screen
    })

    return stimulus_response_xr


def compute_means_xr(stimulus_response_xr, response_window_duration=0.5):
    '''
    Computes means of traces for stimulus response and pre-stimulus baseline.
    Response by default starts at 0, while baseline
    trace by default ends at 0.

    Parameters:
    ___________
    stimulus_response_xr: xarray
        stimulus_response_xr from get_stimulus_response_xr
        with three main dimentions: cell_specimen_id,
        trail_id, and eventlocked_timestamps
    response_window_duration:
        duration in seconds relative to stimulus onset to compute the mean and baseline responses
        in get_stimulus_response_xr

    Returns:
    _________
        stimulus_response_xr with additional
        dimentions: mean_response and mean_baseline
    '''
    response_range = [0, response_window_duration]
    baseline_range = [-response_window_duration, 0]

    mean_response = stimulus_response_xr.loc[
        {'eventlocked_timestamps': slice(*response_range)}
    ].mean(['eventlocked_timestamps'])

    mean_baseline = stimulus_response_xr.loc[
        {'eventlocked_timestamps': slice(*baseline_range)}
    ].mean(['eventlocked_timestamps'])

    stimulus_response_xr = xarray.Dataset({
        'eventlocked_traces': stimulus_response_xr,
        'mean_response': mean_response,
        'mean_baseline': mean_baseline,
    })

    return stimulus_response_xr


def get_spontaneous_frames(stimulus_presentations, ophys_timestamps, gray_screen_period_to_use='before'):
    '''
        Returns a list of the frames that occur during the before and after spontaneous windows. This is copied from VBA. Does not use the full spontaneous period because that is what VBA did. It only uses 4 minutes of the before and after spontaneous period.

    Args:
        stimulus_presentations_df (pandas.DataFrame): table of stimulus presentations, including start_time and stop_time
        ophys_timestamps (np.array): timestamps of each ophys frame
        gray_screen_period_to_use (str): 'before', 'after', or 'both'
                                        whether to use the gray screen period before the session, after the session, or across both
    Returns:
        spontaneous_inds (np.array): indices of ophys frames during the gray screen period before or after the session, or both
    '''
    # exclude the very first minute of the session because the monitor has just turned on and can cause artifacts
    # spont_duration_frames = 4 * 60 * 60  # 4 mins * * 60s/min * 60Hz
    spont_duration = 4 * 60  # 4mins * 60sec

    # for spontaneous at beginning of session, get 4 minutes of gray screen values prior to first stimulus
    if stimulus_presentations.iloc[0].image_name == 'omitted': # something weird happens when first stimulus is omitted, start time is at beginning of session
        first_index = 1
    else:
        first_index = 0
    behavior_start_time = stimulus_presentations.iloc[first_index].start_time
    spontaneous_start_time_pre = behavior_start_time - spont_duration
    spontaneous_end_time_pre = behavior_start_time
    spontaneous_start_frame_pre = general_utilities.index_of_nearest_value(ophys_timestamps, spontaneous_start_time_pre)
    spontaneous_end_frame_pre = general_utilities.index_of_nearest_value(ophys_timestamps, spontaneous_end_time_pre)
    spontaneous_frames_pre = np.arange(spontaneous_start_frame_pre, spontaneous_end_frame_pre, 1)

    # for spontaneous epoch at end of session, get 4 minutes of gray screen values after the last stimulus
    behavior_end_time = stimulus_presentations.iloc[-1].start_time
    spontaneous_start_time_post = behavior_end_time + 0.75
    spontaneous_end_time_post = spontaneous_start_time_post + spont_duration
    spontaneous_start_frame_post = general_utilities.index_of_nearest_value(ophys_timestamps, spontaneous_start_time_post)
    spontaneous_end_frame_post = general_utilities.index_of_nearest_value(ophys_timestamps, spontaneous_end_time_post)
    spontaneous_frames_post = np.arange(spontaneous_start_frame_post, spontaneous_end_frame_post, 1)

    if gray_screen_period_to_use == 'before':
        spontaneous_frames = spontaneous_frames_pre
    elif gray_screen_period_to_use == 'after':
        spontaneous_frames = spontaneous_frames_post
    elif gray_screen_period_to_use == 'both':
        spontaneous_frames = np.concatenate([spontaneous_frames_pre, spontaneous_frames_post])
    return spontaneous_frames



def get_p_value_from_shuffled_spontaneous(mean_responses,
                                      stimulus_presentations,
                                      ophys_timestamps,
                                      traces_array,
                                      response_window_duration,
                                      ophys_frame_rate=None,
                                      number_of_shuffles=10000):
    '''
    Args:
        mean_responses (array): Mean response values, shape (nConditions, nCells)
        stimulus_presentations_df (pandas.DataFrame): Table of stimulus presentations, including start_time and stop_time
        ophys_timestamps (np.array): Timestamps of each ophys frame
        traces_arr (np.array): trace values, shape (nSamples, nCells)
        response_window_duration (int): Number of frames averaged to produce mean response values
        number_of_shuffles (int): Number of shuffles of spontaneous activity used to produce the p-value
    Returns:
        p_values (array): p-value for each response mean, shape (nConditions, nCells)
    '''

    from brain_observatory_utilities.utilities.general_utilities import eventlocked_traces

    spontaneous_frames = get_spontaneous_frames(stimulus_presentations, ophys_timestamps, gray_screen_period_to_use='before')
    shuffled_spont_inds = np.random.choice(spontaneous_frames, number_of_shuffles)

    if ophys_frame_rate is None:
        ophys_frame_rate = 1 / np.diff(ophys_timestamps).mean()

    trace_len = np.round(response_window_duration * ophys_frame_rate).astype(int)
    start_ind_offset = 0
    end_ind_offset = trace_len
    # get an x frame segment of each cells trace after each shuffled spontaneous timepoint
    spont_traces = eventlocked_traces(traces_array, shuffled_spont_inds, start_ind_offset, end_ind_offset)
    # average over the response window (x frames) for each shuffle,
    spont_mean = spont_traces.mean(axis=0)  #Returns (nShuffles, nCells) - mean repsonse following each shuffled spont frame

    # Goal is to figure out how each response compares to the shuffled distribution, which is just
    # a searchsorted call if we first sort the shuffled.
    spont_mean_sorted = np.sort(spont_mean, axis=0) # for each cell, sort the spontaneous mean values (axis0 is shuffles, axis1 is cells)
    response_insertion_ind = np.empty(mean_responses.shape) # should be nConditions, nCells
    # in cases where there is only 1 unique ID (i.e. one neuron in FOV, or one running or pupil trace), duplicate dims so the code below works
    if spont_mean_sorted.ndim == 1:
        spont_mean_sorted = np.expand_dims(spont_mean_sorted, axis=1)
    # loop through indices and figure out how many times the mean response is greater than the spontaneous shuffles
    for ind_cell in range(mean_responses.shape[1]):
        response_insertion_ind[:, ind_cell] = np.searchsorted(spont_mean_sorted[:, ind_cell],
                                                              mean_responses[:, ind_cell])
    # p value is 1 over the fraction times that a given mean response is larger than the 10,000 shuffle means
    # response_insertion_ind tells the index that the mean response would have to be placed in to maintain the order of the shuffled spontaneous
    # if that number is 10k, the mean response is larger than all the shuffles
    # dividing response_insertion_index by 10k gives you the fraction of times that mean response was greater than the shuffles
    # then divide by 1 to get p-value
    proportion_spont_larger_than_sample = 1 - (response_insertion_ind / number_of_shuffles)
    p_values = proportion_spont_larger_than_sample
    # result = xarray.DataArray(data=proportion_spont_larger_than_sample,
    #                       coords=mean_responses.coords)
    return p_values


def get_stimulus_response_df(ophys_experiment,
                             data_type='dff',
                             event_type='all',
                             time_window=[-3, 3],
                             response_window_duration=0.5,
                             interpolate=True,
                             output_sampling_rate=None,
                             exclude_invalid_rois=True,
                             **kwargs):
    '''
    Get stimulus aligned responses from one ophys_experiment.

    Parameters:
    ___________
    ophys_experiment: obj
        AllenSDK BehaviorOphysExperiment object
        A BehaviorOphysExperiment instance
        See https://github.com/AllenInstitute/AllenSDK/blob/master/allensdk/brain_observatory/behavior/behavior_ophys_ophys_experiment.py  # noqa E501
    data_type: str
        neural or behavioral data type to extract, options are: dff (default), events, filtered_events, running_speed, pupil_diameter, lick_rate
    event_type: str
        event type to align to, which can be found in columns of ophys_experiment.stimulus_presentations df.
        options are: 'all' (default) - gets all stimulus trials
                     'images' - gets only image presentations (changes and not changes)
                     'omissions' - gets only trials with omitted stimuli
                     'changes' - get only trials of image changes
    time_window: array
        array of two int or floats indicating the time window on sliced data, default = [-3, 3]
    response_window_duration: float
        time period, in seconds, relative to stimulus onset to compute the mean and baseline response
    interpolate: bool
        type of alignment. If True (default) - interpolates neural data to align timestamps
        with stimulus presentations. If False - shifts data to the nearest timestamps
    output_sampling_rate : float
        Desired sampling of output.
        Input data will be interpolated to this sampling rate if interpolate = True (default). # NOQA E501
        If passing interpolate = False, the sampling rate of the input timeseries will # NOQA E501
        be used and output_sampling_rate should not be specified.
     exclude_invalid_rois : bool
        If True, only ROIs deemed as 'valid' by the classifier will be returned. If False, 'invalid' ROIs will be returned
        This only works if the provided dataset was loaded using internal Allen Institute database, does not work for NWB files.
        In the case that dataset object is loaded through publicly released NWB files, only 'valid' ROIs will be returned

    kwargs: key, value mappings
        Other keyword arguments are passed down to general_utilities.event_triggered_response(),
        for interpolation method such as output_sampling_rate and include_endpoint.

    Returns:
    ___________
    stimulus_response_df: Pandas.DataFrame


    '''

    stimulus_response_xr = get_stimulus_response_xr(
        ophys_experiment=ophys_experiment,
        data_type=data_type,
        event_type=event_type,
        time_window=time_window,
        response_window_duration=response_window_duration,
        interpolate=interpolate,
        output_sampling_rate=output_sampling_rate,
        exclude_invalid_rois=exclude_invalid_rois,
        **kwargs)

    # set up identifier columns depending on whether behavioral or neural data is being used
    # if (data_type == 'running_speed') or (data_type == 'pupil_diameter') or (data_type == 'lick_rate'):
    if ('lick' in data_type) or ('pupil' in data_type) or ('running' in data_type):
        # set up variables to handle only one timeseries per event instead of multiple cell_specimen_ids
        unique_id_string = 'trace_id'  # create a column to take the place of 'cell_specimen_id'
    else:
        # all cell specimen ids in an ophys_experiment
        unique_id_string = 'cell_specimen_id'

    # get mean response after stimulus onset and during pre-stimulus baseline
    mean_response = stimulus_response_xr['mean_response']
    mean_baseline = stimulus_response_xr['mean_baseline']
    stacked_response = mean_response.stack(
        multi_index=('stimulus_presentations_id', unique_id_string)).transpose()  # noqa E501
    stacked_baseline = mean_baseline.stack(
        multi_index=('stimulus_presentations_id', unique_id_string)).transpose()  # noqa E501

    # get p_value for each stimulus response compared to a shuffled distribution of gray screen values
    p_vals_gray_screen = stimulus_response_xr['p_value_gray_screen']
    stacked_pval_gray_screen = p_vals_gray_screen.stack(
        multi_index=('stimulus_presentations_id', unique_id_string)).transpose()  # noqa E501

    # get event locked traces and timestamps from xarray
    traces = stimulus_response_xr['eventlocked_traces']
    stacked_traces = traces.stack(multi_index=(
        'stimulus_presentations_id', unique_id_string)).transpose()
    num_repeats = len(stacked_traces)
    trace_timestamps = np.repeat(
        stacked_traces.coords['eventlocked_timestamps'].data[np.newaxis, :],
        repeats=num_repeats, axis=0)

    # turn it all into a dataframe
    stimulus_response_df = pd.DataFrame({
        'stimulus_presentations_id': stacked_traces.coords['stimulus_presentations_id'],  # noqa E501
        unique_id_string: stacked_traces.coords[unique_id_string],
        'trace': list(stacked_traces.data),
        'trace_timestamps': list(trace_timestamps),
        'mean_response': stacked_response.data,
        'baseline_response': stacked_baseline.data,
        'p_value_gray_screen': stacked_pval_gray_screen,
    })

    # save frame rate, time window and other metadata for reference
    if output_sampling_rate is not None:
        stimulus_response_df['ophys_frame_rate'] = output_sampling_rate
    else:
        stimulus_response_df['ophys_frame_rate'] = ophys_experiment.metadata['ophys_frame_rate']
    stimulus_response_df['data_type'] = data_type
    stimulus_response_df['event_type'] = event_type
    stimulus_response_df['interpolate'] = interpolate
    stimulus_response_df['output_sampling_rate'] = output_sampling_rate
    stimulus_response_df['response_window_duration'] = response_window_duration

    return stimulus_response_df



